# Welcome to Universe!
#
# This file contains the client-side registry of environments.

import logging
import os

# Suppress Twisted's warning about service_identity not being installed.
# We don't need service_identity right now and don't want to take it on
# as a dependency just to suppress this warning.
import warnings
warnings.filterwarnings(
    'ignore',
    message='You do not have a working installation of the service_identity'
)


from gym.envs.registration import register

import universe.scoreboard
import universe.configuration
from universe import error, envs
from universe.remotes import docker_remote
from universe.rewarder import merge_infos
from universe.runtimes.registration import runtime_spec

def docker_image(runtime_id):
    logger.warn('DEPRECATION WARNING: universe.docker_image(runtime_id) is deprecated and will be removed soon. Use runtime_spec(runtime_id).image instead. ')
    return runtime_spec(runtime_id).image


#################### Logging configuration ####################

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

_logging_configured = False

def enable_logfile(path=None):
    raise error.Error('Renamed to "universe.configure_logging()"')

def configure_logging(path=None):
    """
    Set up log levels, and split verbose logs to a file

        Configure the client-side environment logs to print
        to stdout at "info" level, and also to print to a
        verbose log file located at /tmp/universe-<pid>.log
        or another path you specify at "debug" level.
        We suggest calling this method at the beginning of
        your script.
    """

    global _logging_configured
    if _logging_configured:
        return
    _logging_configured = True

    if path is False:
        # Disable logfile
        return
    elif path is None:
        path = '/tmp/universe-{}.log'.format(os.getpid())

    logger.info('Writing logs to file: %s', path)
    # Turn up extra_logger level
    extra_logger.setLevel(logging.DEBUG)
    if path == '-':
        return

    # Add file handler to root logger
    root_logger = logging.getLogger()
    formatter = logging.Formatter('[%(asctime)s] %(message)s')
    handler = logging.FileHandler(path, 'w', encoding='UTF-8')
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)

    # Set extra_logger to *only* use file handler
    extra_logger.propagate = False
    extra_logger.addHandler(handler)

############### Environment registration and runtime specification ###############
#
#    Universe environments are registered with the gym
#    environment registry when the universe module
#    is imported. We use the "tags" field to store
#    additional data specific to Universe.


#------------------------ Gym core environments -----------------------#
#     Asynchronous VNC versions of core gym environments,
#     such as CartPole and Pong

# Note on metadata:
#    Environments send on-screen metadata: the current time, and the
#    time the last action was received from the agent. This timestamp
#    data is used to compute action and observation lags. For core
#    environments, this data is sent using on-screen pixels that encode
#    timestamps.

metadata_pixels = {
    'type': 'pixels',
}

# Should be exactly the same as CartPole-v0
register(
    id='gym-core.CartPoleLowDSync-v0',
    entry_point='universe.wrappers:WrappedGymCoreSyncEnv',
    tags={
        'vnc': True,
        'runtime': 'gym-core',
        'metadata_encoding': metadata_pixels,
    },
    kwargs={
        'rewarder_observation': True,
        'gym_core_id': 'CartPole-v0',
},
    # experience_limit=1000,
    trials=2,
    timestep_limit=500,
)

# Dynamics should match CartPole-v0, but have pixel observations
register(
    id='gym-core.CartPoleSync-v0',
    entry_point='universe.wrappers:WrappedGymCoreSyncEnv',
    tags={
        'vnc': True,
        'runtime': 'gym-core',
        'metadata_encoding': metadata_pixels,
    },
    kwargs={
        'gym_core_id': 'CartPole-v0',
    },
    # experience_limit=1000,
    trials=2,
    timestep_limit=500,
)

# Async cartpole with 4-d observations
register(
    id='gym-core.CartPoleLowD-v0',
    entry_point='universe.wrappers:WrappedGymCoreEnv',
    tags={
        'vnc': True,
        'runtime': 'gym-core',
        'metadata_encoding': metadata_pixels,
    },
    kwargs={
        'rewarder_observation': True,
        'gym_core_id': 'CartPole-v0',
    },
    # experience_limit=1000,
    trials=2,
    timestep_limit=500,
)

register(
    id='gym-core.CartPole-v0',
    entry_point='universe.wrappers:WrappedGymCoreEnv',
    tags={
        'vnc': True,
        'runtime': 'gym-core',
        'metadata_encoding': metadata_pixels,
    },
    kwargs={
        'gym_core_id': 'CartPole-v0',
    },
    # experience_limit=1000,
    trials=2,
    timestep_limit=500,
)

# gym-core.Atari
for game in ['air_raid', 'alien', 'amidar', 'assault', 'asterix',
             'asteroids', 'atlantis', 'bank_heist', 'battle_zone',
             'beam_rider', 'berzerk', 'bowling', 'boxing', 'breakout',
             'carnival', 'centipede', 'chopper_command', 'crazy_climber',
             'demon_attack', 'double_dunk', 'elevator_action', 'enduro',
             'fishing_derby', 'freeway', 'frostbite', 'gopher', 'gravitar',
             'ice_hockey', 'jamesbond', 'journey_escape', 'kangaroo', 'krull',
             'kung_fu_master', 'montezuma_revenge', 'ms_pacman',
             'name_this_game', 'phoenix', 'pitfall', 'pong', 'pooyan',
             'private_eye', 'qbert', 'riverraid', 'road_runner', 'robotank',
             'seaquest', 'skiing', 'solaris', 'space_invaders', 'star_gunner',
             'tennis', 'time_pilot', 'tutankham', 'up_n_down', 'venture',
             'video_pinball', 'wizard_of_wor', 'yars_revenge', 'zaxxon']:
    # space_invaders should yield SpaceInvaders-v0 and SpaceInvaders-ram-v0
    base = ''.join([g.capitalize() for g in game.split('_')]) # SpaceInvaders

    for version in [0, 3]:
        gym_core_id = '{}-v{}'.format(base, version) # e.g. SpaceInvaders-v3
        register(
            id='gym-core.{}'.format(gym_core_id),
            entry_point='universe.wrappers:WrappedGymCoreEnv',
            tags={
                'vnc': True,
                'atari': True,
                'runtime': 'gym-core',
                'metadata_encoding': metadata_pixels,
            },
            kwargs={
                'gym_core_id': gym_core_id,
            },
            # experience_limit=1000,
            timestep_limit=100000,
        )

        register(
            id='gym-core.{}Sync-v{}'.format(base, version),
            entry_point='universe.wrappers:WrappedGymCoreSyncEnv',
            tags={
                'vnc': True,
                'atari': True,
                'runtime': 'gym-core',
                'metadata_encoding': metadata_pixels,
            },
            kwargs={
                'gym_core_id': gym_core_id,
            },
            # experience_limit=1000,
            timestep_limit=100000,
        )

        register(
            id='gym-core.{}30FPS-v{}'.format(base, version),
            entry_point='universe.wrappers:WrappedGymCoreEnv',
            tags={
                'vnc': True,
                'atari': True,
                'runtime': 'gym-core',
                'metadata_encoding': metadata_pixels,
            },
            kwargs={
                'gym_core_id': gym_core_id,
                'fps': 30,
            },
            # experience_limit=1000,
            timestep_limit=100000,
        )

        register(
            id='gym-core.{}Slow-v{}'.format(base, version),
            entry_point='universe.wrappers:WrappedGymCoreEnv',
            tags={
                'vnc': True,
                'atari': True,
                'runtime': 'gym-core',
                'metadata_encoding': metadata_pixels,
            },
            kwargs={
                'gym_core_id': gym_core_id,
                'fps': 15,
            },
            # experience_limit=1000,
            timestep_limit=100000,
        )

        deterministic_gym_core_id = '{}Deterministic-v{}'.format(base, version) # e.g. SpaceInvadersDeterministic-v3

        register(
            id='gym-core.{}Deterministic-v{}'.format(base, version),
            entry_point='universe.wrappers:WrappedGymCoreEnv',
            tags={
                'vnc': True,
                'atari': True,
                'runtime': 'gym-core',
                'metadata_encoding': metadata_pixels,
            },
            kwargs={
                'gym_core_id': deterministic_gym_core_id,
            },
            timestep_limit=100000,
        )
        register(
            id='gym-core.{}DeterministicSlow-v{}'.format(base, version),
            entry_point='universe.wrappers:WrappedGymCoreEnv',
            tags={
                'vnc': True,
                'atari': True,
                'runtime': 'gym-core',
                'metadata_encoding': metadata_pixels,
            },
            kwargs={
                'gym_core_id': deterministic_gym_core_id,
                'fps': 15,
            },
            timestep_limit=75000,
        )
        register(
            id='gym-core.{}DeterministicSync-v{}'.format(base, version),
            entry_point='universe.wrappers:WrappedGymCoreSyncEnv',
            tags={
                'vnc': True,
                'atari': True,
                'runtime': 'gym-core',
                'metadata_encoding': metadata_pixels,
            },
            kwargs={
                'gym_core_id': deterministic_gym_core_id,
            },
            # experience_limit=1000,
            timestep_limit=75000,
        )

        no_frameskip_gym_core_id = '{}NoFrameskip-v{}'.format(base, version) # e.g. SpaceInvadersNoFrameskip-v3
        register(
            id='gym-core.{}NoFrameskip-v{}'.format(base, version),
            entry_point='universe.wrappers:WrappedGymCoreEnv',
            tags={
                'vnc': True,
                'atari': True,
                'runtime': 'gym-core',
                'metadata_encoding': metadata_pixels,
            },
            kwargs={
                'gym_core_id': no_frameskip_gym_core_id,
            },
            # experience_limit=1000,
            timestep_limit=400000
        )

#------------------------ Flash game environments ------------------------#
#     Browser-based flash games, run locally
#     in Chrome within a Docker container

# Note on metadata: flashgames send time metadata using
# an on-screen QR code

metadata_v1 = {
    'type': 'qrcode',
    'x': 914,
    'y': 658,
    'width': 100,
    'height': 100,
}

# Please keep this registry mirrored with the benchmarks in universe-envs/flashgames/gym_flashgames/__init__.py
# You can use universe-envs/flashgames/bin/manage export_env_ids_for_registration to generate this list.
for game in [
    'flashgames.1001ArabianNights-v0',
    'flashgames.21Balloons-v0',
    'flashgames.30Seconds-v0',
    'flashgames.3FootNinja-v0',
    'flashgames.3FootNinjaIi-v0',
    'flashgames.3dBuggyRacing-v0',
    'flashgames.3dClassicRacing-v0',
    'flashgames.3dFlashRacer-v0',
    'flashgames.3dFuriousDriver-v0',
    'flashgames.3dLaSupercars-v0',
    'flashgames.3dLaSupercars2-v0',
    'flashgames.3dMuscleCarRacer-v0',
    'flashgames.3dRallyFever-v0',
    'flashgames.3dRookieCop-v0',
    'flashgames.3dSpeedFever-v0',
    'flashgames.3dSportRampage-v0',
    'flashgames.3dSuperRide-v0',
    'flashgames.3dTestDrive-v0',
    'flashgames.3dTruckInTheWoods-v0',
    'flashgames.3dUrbanMadness-v0',
    'flashgames.3dUrbanMadness2-v0',
    'flashgames.4x4Monster3-v0',
    'flashgames.99BricksTheLegendOfGarry-v0',
    'flashgames.AWeekendAtTweetys-v0',
    'flashgames.AbductionGrannysVersion-v0',
    'flashgames.AchilliaTheGame-v0',
    'flashgames.AcidFactory-v0',
    'flashgames.AdrenalineChaser-v0',
    'flashgames.AdventuresOfBloo-v0',
    'flashgames.AeroDefense-v0',
    'flashgames.AerobaticMaster2-v0',
    'flashgames.Aerorumble-v0',
    'flashgames.AirWar1941-v0',
    'flashgames.AircraftRace-v0',
    'flashgames.AladdinAndTheWonderLamp-v0',
    'flashgames.AlchemySwap-v0',
    'flashgames.AliceNixsAdventure-v0',
    'flashgames.Alien-v0',
    'flashgames.AlienAssault-v0',
    'flashgames.AlienTransporter-v0',
    'flashgames.AmericanRacing-v0',
    'flashgames.AmericanRacing2-v0',
    'flashgames.AmigoPancho-v0',
    'flashgames.AmigoPancho3SheriffSancho-v0',
    'flashgames.AmigoPancho4Travel-v0',
    'flashgames.AmigoPanchoInAfghanistan-v0',
    'flashgames.AngryNewsVan-v0',
    'flashgames.AnimeClicker2-v0',
    'flashgames.AnotherLife2-v0',
    'flashgames.AntsGlider-v0',
    'flashgames.AnywayFish-v0',
    'flashgames.ArkanoidGame-v0',
    'flashgames.ArmyPursuit-v0',
    'flashgames.ArmySpeeder-v0',
    'flashgames.AspenSecret-v0',
    'flashgames.AsphaltMadness-v0',
    'flashgames.AssembleBots-v0',
    'flashgames.Astroman-v0',
    'flashgames.AtvRide-v0',
    'flashgames.Autoattack-v0',
    'flashgames.Avalancher-v0',
    'flashgames.AwesomeRun2-v0',
    'flashgames.BackHome-v0',
    'flashgames.BaldEagleJigsawPuzzle-v0',
    'flashgames.BalloonGods-v0',
    'flashgames.BalloonHero-v0',
    'flashgames.BalloonsPop-v0',
    'flashgames.Basement-v0',
    'flashgames.BeachCrazy-v0',
    'flashgames.BearInSuperActionAdventure-v0',
    'flashgames.BigWheelsTrial-v0',
    'flashgames.BikeTrial-v0',
    'flashgames.BikeTrial2-v0',
    'flashgames.BikeTrial3-v0',
    'flashgames.BikeTrial4-v0',
    'flashgames.Bimmin2-v0',
    'flashgames.BirdSlice-v0',
    'flashgames.BirdsFeeding-v0',
    'flashgames.BlackAndWhiteEscapeTheOffice-v0',
    'flashgames.BlackForce-v0',
    'flashgames.BlackInk-v0',
    'flashgames.BlackRacerJigsawPuzzle-v0',
    'flashgames.BlacksmithLab-v0',
    'flashgames.BlastTheMooks-v0',
    'flashgames.BlastTheMooksLevelPack-v0',
    'flashgames.Blix-v0',
    'flashgames.BlobsStory-v0',
    'flashgames.BlockysEscape-v0',
    'flashgames.BloodbathBay-v0',
    'flashgames.BloodyMonstersPack2-v0',
    'flashgames.Blosics2-v0',
    'flashgames.Blosics2LevelPack-v0',
    'flashgames.Blosics3-v0',
    'flashgames.BoatDrive-v0',
    'flashgames.BobbyNutcaseMotoJumping-v0',
    'flashgames.BoltThrough-v0',
    'flashgames.BombIt4-v0',
    'flashgames.BombIt6-v0',
    'flashgames.BombThePiratePigs-v0',
    'flashgames.BottleCaps-v0',
    'flashgames.BouncyCannon-v0',
    'flashgames.BoxBlocks-v0',
    'flashgames.BoxRacers-v0',
    'flashgames.BoxingLiveRound2-v0',
    'flashgames.BraveAstronaut-v0',
    'flashgames.BraveHeads-v0',
    'flashgames.BubbleAdventures-v0',
    'flashgames.BubbleBlubbs-v0',
    'flashgames.BubbleGlee-v0',
    'flashgames.BubbleHit-v0',
    'flashgames.BubbleHitChristmas-v0',
    'flashgames.BubbleHitHalloween-v0',
    'flashgames.BubbleHitPonyParade-v0',
    'flashgames.BubbleHitValentine-v0',
    'flashgames.BubbleMover-v0',
    'flashgames.BubblePop-v0',
    'flashgames.BubblePopAdventure-v0',
    'flashgames.BubbleRubble-v0',
    'flashgames.BubbleRubbleTheIsland-v0',
    'flashgames.BubbleShooterChallenge-v0',
    'flashgames.BubbleSlasher-v0',
    'flashgames.BubbleTanksTd15-v0',
    'flashgames.BubbleTub-v0',
    'flashgames.BubblesInSpace-v0',
    'flashgames.BugsGotGuns-v0',
    'flashgames.BuildBalance2-v0',
    'flashgames.BulletFury-v0',
    'flashgames.BulletHeaven-v0',
    'flashgames.BulletHeaven2-v0',
    'flashgames.Bullets-v0',
    'flashgames.BullfrogJigsawPuzzle-v0',
    'flashgames.BumbleTumble-v0',
    'flashgames.BunnyAndSquirt-v0',
    'flashgames.BunnyCannon-v0',
    'flashgames.BurgerBar-v0',
    'flashgames.BushRoyalRampage-v0',
    'flashgames.Business-v0',
    'flashgames.BusinessmanSimulator-v0',
    'flashgames.CableCapers2-v0',
    'flashgames.CakeQuest-v0',
    'flashgames.CandyMatch-v0',
    'flashgames.CandyMatchCrush-v0',
    'flashgames.CandySlider-v0',
    'flashgames.Canopy-v0',
    'flashgames.CanyonValleyRally-v0',
    'flashgames.CaptainNutty-v0',
    'flashgames.CaptainSteelbounce-v0',
    'flashgames.CardinalQuest2-v0',
    'flashgames.CarrotFantasy-v0',
    'flashgames.CarrotFantasy2Desert-v0',
    'flashgames.CarrotFantasy2Undersea-v0',
    'flashgames.CarrotFantasyExtreme2-v0',
    'flashgames.CarrotFantasyExtreme3-v0',
    'flashgames.CarsVsRobots-v0',
    'flashgames.CartoonCandy-v0',
    'flashgames.CastleRush-v0',
    'flashgames.CastleSolitaire-v0',
    'flashgames.CatGodVsSunKing2-v0',
    'flashgames.CatchTheStar-v0',
    'flashgames.Cattlepult-v0',
    'flashgames.CattlepultPlayerPack-v0',
    'flashgames.CavemanEscape-v0',
    'flashgames.CemeteryRoad-v0',
    'flashgames.CharlieTheDuck-v0',
    'flashgames.Chefday-v0',
    'flashgames.ChickCannont-v0',
    'flashgames.ChickInduce-v0',
    'flashgames.ChockABox-v0',
    'flashgames.ChristmasBubbles-v0',
    'flashgames.ChromaticTowerDefense-v0',
    'flashgames.ChuteAcademy-v0',
    'flashgames.CircuitSuperCarsRacing-v0',
    'flashgames.CitySiege3FubarLevelPack-v0',
    'flashgames.CitySkyTyping-v0',
    'flashgames.ClaustrophobiumFourStepsFromDeath-v0',
    'flashgames.Cleopatra-v0',
    'flashgames.ClickerMonsters-v0',
    'flashgames.ClimbOrDrown2-v0',
    'flashgames.ClimberGuy-v0',
    'flashgames.ClimbingSanta-v0',
    'flashgames.CloseCombat-v0',
    'flashgames.Cloud9-v0',
    'flashgames.ClubNitro-v0',
    'flashgames.Clusterobot-v0',
    'flashgames.CoastRunners-v0',
    'flashgames.CoasterCars2Contact-v0',
    'flashgames.CoasterCars2Megacross-v0',
    'flashgames.CoasterCarsBridgesTrack-v0',
    'flashgames.CoasterCarsCJackTrack-v0',
    'flashgames.CoasterRacer-v0',
    'flashgames.CoasterRacerLvl2-v0',
    'flashgames.CoasterRacerLvl3-v0',
    'flashgames.CoasterRacerLvl4-v0',
    'flashgames.CoasterRacerLvl5-v0',
    'flashgames.CoasterRacerLvl6-v0',
    'flashgames.CoasterRacerLvl7-v0',
    'flashgames.CoasterRacerLvl8-v0',
    'flashgames.CoasterRacer2-v0',
    'flashgames.CoasterRacer2Bike-v0',
    'flashgames.CoasterRacer3-v0',
    'flashgames.CoffeeClicker-v0',
    'flashgames.ColorZapper-v0',
    'flashgames.Colordefense-v0',
    'flashgames.Colorfill-v0',
    'flashgames.Coloruid-v0',
    'flashgames.Colorwars-v0',
    'flashgames.Commando-v0',
    'flashgames.Commando2-v0',
    'flashgames.Conjure-v0',
    'flashgames.Connect2-v0',
    'flashgames.Conquerium-v0',
    'flashgames.Contra3TheAlienWars-v0',
    'flashgames.Cooliobeat-v0',
    'flashgames.Cooliodj-v0',
    'flashgames.CopperheadJigsawPuzzle-v0',
    'flashgames.CosmicSwitch-v0',
    'flashgames.CosmoGravity2-v0',
    'flashgames.CoverOrangeJourneyGangsters-v0',
    'flashgames.CowboyVsUfos-v0',
    'flashgames.Crane-v0',
    'flashgames.CrapImBroke-v0',
    'flashgames.CrazyDarts-v0',
    'flashgames.Crazycle-v0',
    'flashgames.CruiseAdventure-v0',
    'flashgames.Cruisin-v0',
    'flashgames.Crumbs2-v0',
    'flashgames.CrystalCurse-v0',
    'flashgames.CrystalStoryIi-v0',
    'flashgames.CupidBubbles-v0',
    'flashgames.CursedTreasureDontTouchMyGems-v0',
    'flashgames.CurveFever-v0',
    'flashgames.DaleAndPeakot-v0',
    'flashgames.DanceBattle-v0',
    'flashgames.DancingWithShadows-v0',
    'flashgames.DartsSim-v0',
    'flashgames.DaymareInvaders-v0',
    'flashgames.DeadHungry2-v0',
    'flashgames.DeathCabin-v0',
    'flashgames.DeathDiceOverdose-v0',
    'flashgames.DeepForest3dRace-v0',
    'flashgames.DeepFreeze-v0',
    'flashgames.Deliveryman-v0',
    'flashgames.DetectiveConrad-v0',
    'flashgames.Devilment-v0',
    'flashgames.DiamondCrashMania-v0',
    'flashgames.DigToChina-v0',
    'flashgames.DinoBubble-v0',
    'flashgames.DinoMeatHunt3Extra-v0',
    'flashgames.DirkValentine-v0',
    'flashgames.DisasterWillStrikeDefender-v0',
    'flashgames.DisasterWillStrikeUltimateDisaster-v0',
    'flashgames.DiscoverEurope-v0',
    'flashgames.Ditloid-v0',
    'flashgames.DnaLabRush-v0',
    'flashgames.Dodge-v0',
    'flashgames.DodgeAndCrash-v0',
    'flashgames.DolphinVolleyball-v0',
    'flashgames.DontPanic-v0',
    'flashgames.DoodleGod2Walkthrough-v0',
    'flashgames.DotGrowth-v0',
    'flashgames.Dots-v0',
    'flashgames.DotsRevamped-v0',
    'flashgames.DoubleEdged-v0',
    'flashgames.DoughSnake-v0',
    'flashgames.DragonChain-v0',
    'flashgames.DragonChronicles-v0',
    'flashgames.DragonFortress-v0',
    'flashgames.DragonFunflap-v0',
    'flashgames.DragonVsMonster-v0',
    'flashgames.DrawGems-v0',
    'flashgames.DreamChristmasLink-v0',
    'flashgames.DriftRunners-v0',
    'flashgames.DriftRunners2-v0',
    'flashgames.DriftRunners3d-v0',
    'flashgames.Drifters-v0',
    'flashgames.DrinkBeerNeglectFamily-v0',
    'flashgames.DriveToWreck-v0',
    'flashgames.DriveToWreck2-v0',
    'flashgames.DriveToWreck3-v0',
    'flashgames.Dropblox-v0',
    'flashgames.DualDimension-v0',
    'flashgames.DumperRush-v0',
    'flashgames.DungeonBlocks-v0',
    'flashgames.DuskDrive-v0',
    'flashgames.DuskRacers-v0',
    'flashgames.EasterBubbles-v0',
    'flashgames.EasterBunnyCollectCarrots-v0',
    'flashgames.EasterBunnyEggs-v0',
    'flashgames.EasterEggSlider-v0',
    'flashgames.EasterEggsChallenge-v0',
    'flashgames.EatToWin-v0',
    'flashgames.EctoHarvest-v0',
    'flashgames.EffingWorms-v0',
    'flashgames.EggzBlast-v0',
    'flashgames.EiffelTowerAtNight-v0',
    'flashgames.ElClassico-v0',
    'flashgames.ElainesBakery-v0',
    'flashgames.EmpireBusiness2Beta-v0',
    'flashgames.Enhanced-v0',
    'flashgames.EpicBattleFantasy4-v0',
    'flashgames.EpicDefender-v0',
    'flashgames.EpicDerbyRace-v0',
    'flashgames.EpicTimePirates-v0',
    'flashgames.EscapeTheRedGiant-v0',
    'flashgames.EuroKicks2016-v0',
    'flashgames.EvasiveRacers-v0',
    'flashgames.EvilMinion-v0',
    'flashgames.EvilSun-v0',
    'flashgames.EvolutionRacing-v0',
    'flashgames.ExperimentalShooter2-v0',
    'flashgames.ExploreTheCandies-v0',
    'flashgames.ExtremeAirWars-v0',
    'flashgames.ExtremeSkiing-v0',
    'flashgames.F1RacingChallenge-v0',
    'flashgames.FairyDefense-v0',
    'flashgames.FallDamage-v0',
    'flashgames.FarmRush-v0',
    'flashgames.FasterMiterMaster-v0',
    'flashgames.FeedMeMoar-v0',
    'flashgames.FeedOurDoughnutOverlords-v0',
    'flashgames.Filler-v0',
    'flashgames.Filler2-v0',
    'flashgames.FinalNinjaZero-v0',
    'flashgames.FinalSiege-v0',
    'flashgames.FindTheCandy3Kids-v0',
    'flashgames.Firebug-v0',
    'flashgames.FirefighterCannon-v0',
    'flashgames.FireworksGame-v0',
    'flashgames.FishAndDestroy-v0',
    'flashgames.FishEatFish-v0',
    'flashgames.FitItQuick-v0',
    'flashgames.FiveTil-v0',
    'flashgames.Fizzion-v0',
    'flashgames.Flagman-v0',
    'flashgames.FlappyAdventure-v0',
    'flashgames.FlappyBat-v0',
    'flashgames.FlappyCopter-v0',
    'flashgames.FlappyPanda-v0',
    'flashgames.FlashBombs-v0',
    'flashgames.FlashDrive-v0',
    'flashgames.FlashRace-v0',
    'flashgames.FlashRacer-v0',
    'flashgames.Flashcycle2-v0',
    'flashgames.FlashsBounty-v0',
    'flashgames.FlowerGuardian-v0',
    'flashgames.FlowerSolitaire-v0',
    'flashgames.FluffRush-v0',
    'flashgames.FlyAwayRabbit2-v0',
    'flashgames.FlyPlane-v0',
    'flashgames.FlyingCookieQuest-v0',
    'flashgames.FlyingKiwi-v0',
    'flashgames.FlyingTest-v0',
    'flashgames.Foosball2Player-v0',
    'flashgames.FootballHeads201314Ligue1-v0',
    'flashgames.FormulaRacer-v0',
    'flashgames.FormulaRacer2012-v0',
    'flashgames.FormulaXspeed3d-v0',
    'flashgames.FoxSnakeJigsawPuzzle-v0',
    'flashgames.FpaWorld1Remix-v0',
    'flashgames.FreakyRun-v0',
    'flashgames.FredFigglehorn-v0',
    'flashgames.FreeSouls-v0',
    'flashgames.Free_to_use-v0',
    'flashgames.FreecellDuplex-v0',
    'flashgames.FrogEatFlies-v0',
    'flashgames.Frogged-v0',
    'flashgames.FrozenImps-v0',
    'flashgames.FrozenIslandsNewHorizons-v0',
    'flashgames.Funkostroll-v0',
    'flashgames.FunnyEaster-v0',
    'flashgames.GSwitch-v0',
    'flashgames.GalacticCats-v0',
    'flashgames.GalacticGems-v0',
    'flashgames.GalacticGems2-v0',
    'flashgames.GalacticGems2Accelerated-v0',
    'flashgames.GalacticGems2LevelPack-v0',
    'flashgames.GalacticGems2NewFrontiers-v0',
    'flashgames.GalaxyDefender-v0',
    'flashgames.GalaxyEvo2-v0',
    'flashgames.GalaxyMission-v0',
    'flashgames.GalleonFight-v0',
    'flashgames.GameInit-v0',
    'flashgames.Gameinit-v0',
    'flashgames.GamerMemoryTest-v0',
    'flashgames.GardenRush-v0',
    'flashgames.GasSand-v0',
    'flashgames.GemMania-v0',
    'flashgames.GemPop-v0',
    'flashgames.Gemclix-v0',
    'flashgames.Gemcraft-v0',
    'flashgames.GemstoneCastle-v0',
    'flashgames.GhostClimb2Player-v0',
    'flashgames.GiantsAndDwarvesTd-v0',
    'flashgames.GlobalRallyRacer-v0',
    'flashgames.Gloom-v0',
    'flashgames.Gluey2-v0',
    'flashgames.Go-v0',
    'flashgames.GoGreenGo-v0',
    'flashgames.GoKart3d-v0',
    'flashgames.Goldextraction-v0',
    'flashgames.GolfRun-v0',
    'flashgames.GonAndMon-v0',
    'flashgames.GrandPrixGo-v0',
    'flashgames.GrandPrixGo2-v0',
    'flashgames.GrappleCat-v0',
    'flashgames.GravityBall-v0',
    'flashgames.GravityGuy-v0',
    'flashgames.GravityThruster-v0',
    'flashgames.GroundBattles-v0',
    'flashgames.Growbox-v0',
    'flashgames.GsSoccerWorldCup-v0',
    'flashgames.GunExpress-v0',
    'flashgames.GunnerMayhem-v0',
    'flashgames.GunpowderAndFeathers-v0',
    'flashgames.HalloweenAdventureRun-v0',
    'flashgames.HalloweenExplorer-v0',
    'flashgames.HalloweenJam-v0',
    'flashgames.HammerBall-v0',
    'flashgames.Hamsterball-v0',
    'flashgames.HandsOff-v0',
    'flashgames.HappyBallz-v0',
    'flashgames.HappyBees-v0',
    'flashgames.HappyEasterEggs-v0',
    'flashgames.Harvest-v0',
    'flashgames.HarvestDay-v0',
    'flashgames.Hash-v0',
    'flashgames.Hearts-v0',
    'flashgames.HeatRushFuture-v0',
    'flashgames.HeatRushFutureLvl2-v0',
    'flashgames.HeatRushFutureLvl3-v0',
    'flashgames.HeatRushFutureLvl4-v0',
    'flashgames.HeatRushFutureLvl5-v0',
    'flashgames.HeatRushFutureLvl6-v0',
    'flashgames.HeatRushFutureLvl7-v0',
    'flashgames.HeatRushFutureLvl8-v0',
    'flashgames.HeatRushFutureLvl9-v0',
    'flashgames.HeatRushFutureLvl10-v0',
    'flashgames.HeatRushFutureLvl11-v0',
    'flashgames.HeatRushFutureLvl12-v0',
    'flashgames.HeatRushFutureLvl13-v0',
    'flashgames.HeatRushFutureLvl14-v0',
    'flashgames.HeatRushFutureLvl15-v0',
    'flashgames.HeatRushUsa-v0',
    'flashgames.HeatRushUsaLvl2-v0',
    'flashgames.HeatRushUsaLvl3-v0',
    'flashgames.HeatRushUsaLvl4-v0',
    'flashgames.HeatRushUsaLvl5-v0',
    'flashgames.HeatRushUsaLvl6-v0',
    'flashgames.HeatRushUsaLvl7-v0',
    'flashgames.HeatRushUsaLvl8-v0',
    'flashgames.HeatRushUsaLvl9-v0',
    'flashgames.HeatRushUsaLvl10-v0',
    'flashgames.HeatRushUsaLvl11-v0',
    'flashgames.HeatRushUsaLvl12-v0',
    'flashgames.HeatRushUsaLvl13-v0',
    'flashgames.HeatRushUsaLvl14-v0',
    'flashgames.HeatRushUsaLvl15-v0',
    'flashgames.HeatRushUsaLvl16-v0',
    'flashgames.HeavenAndHell-v0',
    'flashgames.HeavyLegion2-v0',
    'flashgames.HeliVsTower-v0',
    'flashgames.HelicopsTerritories-v0',
    'flashgames.Helicrane-v0',
    'flashgames.Helixteus-v0',
    'flashgames.HelmetBombers3-v0',
    'flashgames.HeroRoofTop-v0',
    'flashgames.HeroSimulator-v0',
    'flashgames.HeroesOfMangaraTheFrostCrown-v0',
    'flashgames.HexBattles-v0',
    'flashgames.HeySummer-v0',
    'flashgames.HighSpeedChase-v0',
    'flashgames.HighwayRevenge-v0',
    'flashgames.HiredHeroes-v0',
    'flashgames.HoldTheFort-v0',
    'flashgames.HoleInOne-v0',
    'flashgames.Hotspot-v0',
    'flashgames.HowDareYou-v0',
    'flashgames.HungerHunter-v0',
    'flashgames.HungryLittlePenguins-v0',
    'flashgames.HungryPiranha-v0',
    'flashgames.HunterForDismantlers-v0',
    'flashgames.HyperTravel-v0',
    'flashgames.IceBlock-v0',
    'flashgames.IceCreamFromSpace-v0',
    'flashgames.IceRun-v0',
    'flashgames.IceSlide-v0',
    'flashgames.Ics2-v0',
    'flashgames.IcyGifts2-v0',
    'flashgames.IdleChop-v0',
    'flashgames.IdleFarmer-v0',
    'flashgames.IdleLifting-v0',
    'flashgames.IdlePlanet-v0',
    'flashgames.ImitationNationSnakeGame-v0',
    'flashgames.IncrementalAcceleration-v0',
    'flashgames.Indefinite-v0',
    'flashgames.IndependenceDaySlacking2015-v0',
    'flashgames.InfectonatorSurvivorsAlphaDemo-v0',
    'flashgames.InfernalMess-v0',
    'flashgames.Infinitix-v0',
    'flashgames.InsaneCircle-v0',
    'flashgames.IntoSpace-v0',
    'flashgames.IslandDefense-v0',
    'flashgames.IsoblockerMaster-v0',
    'flashgames.ItsDarkInHell-v0',
    'flashgames.JakeTheSnake-v0',
    'flashgames.JamesTheCircusZebra-v0',
    'flashgames.JamesTheDeepSeaZebra-v0',
    'flashgames.JamesThePirateZebra-v0',
    'flashgames.JamesTheSpaceZebra-v0',
    'flashgames.JelliesFun-v0',
    'flashgames.JellyFriend-v0',
    'flashgames.JellySnake-v0',
    'flashgames.JetpackJackride-v0',
    'flashgames.JollySwipe-v0',
    'flashgames.JollySwipeLevelPack-v0',
    'flashgames.JonnyBackflip-v0',
    'flashgames.JumpOverTheRings-v0',
    'flashgames.Jumprunner-v0',
    'flashgames.Jumpz-v0',
    'flashgames.JungleCrash-v0',
    'flashgames.JungleEagle-v0',
    'flashgames.KamikazeRace-v0',
    'flashgames.KangoIslands-v0',
    'flashgames.KartOn-v0',
    'flashgames.KartRacing-v0',
    'flashgames.KartingSuperGo-v0',
    'flashgames.Kawairun-v0',
    'flashgames.KeeperOfTheGrove3-v0',
    'flashgames.Kinetikz-v0',
    'flashgames.Kinetikz2-v0',
    'flashgames.Kinetikz3-v0',
    'flashgames.KingRolla-v0',
    'flashgames.KitchenRestaurantCleanUp-v0',
    'flashgames.KnightsOfRock-v0',
    'flashgames.Knighttron-v0',
    'flashgames.Knockers-v0',
    'flashgames.Krome-v0',
    'flashgames.LaserCannon3LevelsPack-v0',
    'flashgames.LawnmowerRacing3d-v0',
    'flashgames.LaxAirbusParking-v0',
    'flashgames.Lazerman-v0',
    'flashgames.LearnToFlyIdle-v0',
    'flashgames.Legor9-v0',
    'flashgames.LessQuick-v0',
    'flashgames.LetsFall-v0',
    'flashgames.LevelEditor3-v0',
    'flashgames.LilyFighters-v0',
    'flashgames.LineGameLimeEdition-v0',
    'flashgames.LlamasInDistress-v0',
    'flashgames.LonelyEscapeAsylum-v0',
    'flashgames.LongJump-v0',
    'flashgames.Long_short-v0',
    'flashgames.LooneyAndJohny-v0',
    'flashgames.LuckyBalls-v0',
    'flashgames.LuxUltimate-v0',
    'flashgames.Madburger3-v0',
    'flashgames.MadpetSkateboarder2-v0',
    'flashgames.MagicSafari-v0',
    'flashgames.ManicRallyGo-v0',
    'flashgames.MapTurtleJigsawPuzzle-v0',
    'flashgames.MarblesShooter-v0',
    'flashgames.MarsColonyTd-v0',
    'flashgames.MarshmallowsEscape-v0',
    'flashgames.MashaCollectsButterflies-v0',
    'flashgames.MasterDifference-v0',
    'flashgames.Match2Collapse-v0',
    'flashgames.Match3Adventure-v0',
    'flashgames.Match3ChristmasPack-v0',
    'flashgames.Match3PresentBoxSaga-v0',
    'flashgames.MatchAndCrash-v0',
    'flashgames.MatchAndSpell-v0',
    'flashgames.MatchAroundTheWorld-v0',
    'flashgames.MatchCraft-v0',
    'flashgames.MatchCrypt-v0',
    'flashgames.MatchJong-v0',
    'flashgames.MatchMonsters-v0',
    'flashgames.MatchStars-v0',
    'flashgames.MatchTheBugz-v0',
    'flashgames.MatchTheFruits-v0',
    'flashgames.MatchToEnjoy-v0',
    'flashgames.MatchToEnjoyLevelPack-v0',
    'flashgames.MatchingSweetHearts-v0',
    'flashgames.MazeEye-v0',
    'flashgames.MedievalShark-v0',
    'flashgames.MeerkatMission-v0',
    'flashgames.MexicoRex-v0',
    'flashgames.MiceVsHammers-v0',
    'flashgames.Michimind-v0',
    'flashgames.MidnightCanine-v0',
    'flashgames.MidnightMiner-v0',
    'flashgames.MightyTower-v0',
    'flashgames.Mimelet-v0',
    'flashgames.MindImpulse-v0',
    'flashgames.MineDrop-v0',
    'flashgames.MineHero-v0',
    'flashgames.MinedigJourneyToHollowEarth-v0',
    'flashgames.MiniMachines-v0',
    'flashgames.MiniSportsChallenge-v0',
    'flashgames.MinicarHunt-v0',
    'flashgames.Minicarting-v0',
    'flashgames.MissionEscapeTheDojo-v0',
    'flashgames.ModelCarRacing-v0',
    'flashgames.MonkeyBlast-v0',
    'flashgames.MonkeyGems-v0',
    'flashgames.MonkeyGoHappyNinjaHunt2-v0',
    'flashgames.MonkeyManic-v0',
    'flashgames.MonsterChains-v0',
    'flashgames.MonsterLabFeedThemAll-v0',
    'flashgames.MonsterRun-v0',
    'flashgames.MonsterTroubles-v0',
    'flashgames.MonsterTruckFever-v0',
    'flashgames.MonsterTruckRally-v0',
    'flashgames.Moosters-v0',
    'flashgames.MotherLoad-v0',
    'flashgames.MotoMadness-v0',
    'flashgames.MotoTrialMania-v0',
    'flashgames.MotorWheels-v0',
    'flashgames.Mrbirdie-v0',
    'flashgames.MultiballMadness-v0',
    'flashgames.Multitask-v0',
    'flashgames.MummyMadness-v0',
    'flashgames.Mushbooms-v0',
    'flashgames.MushboomsLevelPack-v0',
    'flashgames.MushboomsLevelPack2-v0',
    'flashgames.MushroomFarmDefender-v0',
    'flashgames.MushyMishy-v0',
    'flashgames.MusicSmash-v0',
    'flashgames.MusicStomp-v0',
    'flashgames.MusicZap-v0',
    'flashgames.MysteriousPirateJewels-v0',
    'flashgames.MysticIndiaPop-v0',
    'flashgames.MysticalAncientTreasure-v0',
    'flashgames.NOfficialWebVersion-v0',
    'flashgames.NadiasRage-v0',
    'flashgames.NanoKingdoms2JokersRevenge-v0',
    'flashgames.NeonRace-v0',
    'flashgames.NeonRaceLvl2-v0',
    'flashgames.NeonRaceLvl3-v0',
    'flashgames.NeonRaceLvl4-v0',
    'flashgames.NeonRaceLvl5-v0',
    'flashgames.NeonRaceLvl6-v0',
    'flashgames.NeonRaceLvl7-v0',
    'flashgames.NeonRaceLvl8-v0',
    'flashgames.NeonRace2-v0',
    'flashgames.Neopods-v0',
    'flashgames.NervousLadybug-v0',
    'flashgames.NewSiberianSupercarsRacing-v0',
    'flashgames.NewSplitterPals-v0',
    'flashgames.NightDrivin-v0',
    'flashgames.NightRaceRally-v0',
    'flashgames.NinjaPainter-v0',
    'flashgames.NinjaPandaArena-v0',
    'flashgames.NinjaPandaCouple-v0',
    'flashgames.NinjaTrainingWorlds-v0',
    'flashgames.Nook-v0',
    'flashgames.NoughtsAndCrosses-v0',
    'flashgames.NoughtsAndCrossesExtreme-v0',
    'flashgames.NukeDefense-v0',
    'flashgames.Numz-v0',
    'flashgames.NuttyBoom-v0',
    'flashgames.ObamaAlienDefense-v0',
    'flashgames.OceanMatch-v0',
    'flashgames.Oddball2-v0',
    'flashgames.OffRoaders3d-v0',
    'flashgames.OfficeTrap-v0',
    'flashgames.Offroaders-v0',
    'flashgames.Offroaders2-v0',
    'flashgames.OkParking-v0',
    'flashgames.OldTv-v0',
    'flashgames.OozingForever-v0',
    'flashgames.OswaldTheAngryDwarf-v0',
    'flashgames.Overheat-v0',
    'flashgames.PaintWars-v0',
    'flashgames.Paintwars-v0',
    'flashgames.PanikInChocoland-v0',
    'flashgames.PapaLouie3WhenSundaesAttack-v0',
    'flashgames.PaperDefense-v0',
    'flashgames.ParachuteRetrospect-v0',
    'flashgames.ParallelLevels-v0',
    'flashgames.ParkingFury-v0',
    'flashgames.Parkour-v0',
    'flashgames.ParticleWarsExtreme-v0',
    'flashgames.Pathillogical-v0',
    'flashgames.PaulVaulting-v0',
    'flashgames.Peakart-v0',
    'flashgames.PearlBreaking-v0',
    'flashgames.Pel-v0',
    'flashgames.PenguinCubes-v0',
    'flashgames.PenguinHeroes-v0',
    'flashgames.PenguinSkate2-v0',
    'flashgames.PerilousJourney2-v0',
    'flashgames.Phit-v0',
    'flashgames.PickAndDig2-v0',
    'flashgames.PickUpTruckRacing-v0',
    'flashgames.PicnicPanicTd-v0',
    'flashgames.PigDestroyer-v0',
    'flashgames.PiggyWiggy-v0',
    'flashgames.PiggysCupcakeQuest-v0',
    'flashgames.PinBalls-v0',
    'flashgames.PinataWarriors-v0',
    'flashgames.PingPongSurvival-v0',
    'flashgames.PirateRunAway-v0',
    'flashgames.PiratesAndCannons-v0',
    'flashgames.PixelBasher-v0',
    'flashgames.PixelFighta-v0',
    'flashgames.PixelPurge-v0',
    'flashgames.PixelQuest-v0',
    'flashgames.PlaneRace-v0',
    'flashgames.PlaneRace2-v0',
    'flashgames.PlopPlopLite-v0',
    'flashgames.PocketRocket-v0',
    'flashgames.Pointer-v0',
    'flashgames.Pointless-v0',
    'flashgames.PoliceChaseCrackdown-v0',
    'flashgames.PoliceHotRacing-v0',
    'flashgames.PoliceInterceptor-v0',
    'flashgames.PolygonalFury-v0',
    'flashgames.Popopop-v0',
    'flashgames.Popopop2-v0',
    'flashgames.PouJetpack-v0',
    'flashgames.PouThanksgivingDaySlacking-v0',
    'flashgames.PowerCopter-v0',
    'flashgames.PowerSwing-v0',
    'flashgames.Primary-v0',
    'flashgames.PrincessBubblesRescuePrince-v0',
    'flashgames.PrincessToTheRescue-v0',
    'flashgames.ProjectMonochrome-v0',
    'flashgames.PuddingPie-v0',
    'flashgames.PufferFish-v0',
    'flashgames.PumpkinCollector-v0',
    'flashgames.PumpkinMan-v0',
    'flashgames.PumpkinsInZombieTown-v0',
    'flashgames.PunchBallJump-v0',
    'flashgames.PurifyTheLegendOfZ-v0',
    'flashgames.PuzzleMonsters-v0',
    'flashgames.PuzzleRescuePrime-v0',
    'flashgames.PyramidApocalypse-v0',
    'flashgames.Pyro-v0',
    'flashgames.Qoosh-v0',
    'flashgames.QuashBoard-v0',
    'flashgames.QubedMysteriousIsland-v0',
    'flashgames.QubeyTheCube-v0',
    'flashgames.Quick-v0',
    'flashgames.RabbitPlanetEscape-v0',
    'flashgames.RabbitRustler-v0',
    'flashgames.RacerKartz-v0',
    'flashgames.RacingSupercarChampionship-v0',
    'flashgames.RainbowDrops-v0',
    'flashgames.RapaNui-v0',
    'flashgames.Raze3-v0',
    'flashgames.Rb2-v0',
    'flashgames.Rbots-v0',
    'flashgames.ReachTheGoal-v0',
    'flashgames.RedBeard-v0',
    'flashgames.RedCode3-v0',
    'flashgames.RedFuryRacing-v0',
    'flashgames.ReleaseTheMooks-v0',
    'flashgames.ReleaseTheMooks2-v0',
    'flashgames.ReleaseTheMooks3-v0',
    'flashgames.Resonance-v0',
    'flashgames.RetroRunner-v0',
    'flashgames.Retron-v0',
    'flashgames.ReverseBoots-v0',
    'flashgames.RhythmBlasterV2-v0',
    'flashgames.RhythmRockets-v0',
    'flashgames.RhythmSnake-v0',
    'flashgames.RingsideHero-v0',
    'flashgames.RiseOfChampions-v0',
    'flashgames.RoadRacing-v0',
    'flashgames.RoadblockAttack-v0',
    'flashgames.RoboPop-v0',
    'flashgames.RobotDuelFight-v0',
    'flashgames.RobotWantsFishy-v0',
    'flashgames.RocketBootsInc-v0',
    'flashgames.Rocketeer-v0',
    'flashgames.RollTheCluster-v0',
    'flashgames.RollerRider-v0',
    'flashgames.RollingHills-v0',
    'flashgames.Rose-v0',
    'flashgames.RubbleRacer-v0',
    'flashgames.RunFaustoRun-v0',
    'flashgames.RunNGun-v0',
    'flashgames.RunRamRun-v0',
    'flashgames.RunRunRan-v0',
    'flashgames.RunSoldierRun-v0',
    'flashgames.RushOfTanks-v0',
    'flashgames.RussianTruck-v0',
    'flashgames.SafariTime-v0',
    'flashgames.SandcastleShowdown-v0',
    'flashgames.SantaClimbHere-v0',
    'flashgames.SantaMan-v0',
    'flashgames.SantaSituation-v0',
    'flashgames.SapphireClix-v0',
    'flashgames.SaveTheDummyHolidays-v0',
    'flashgames.SavingLittleAlien-v0',
    'flashgames.SchoolBusRacing-v0',
    'flashgames.Scribble-v0',
    'flashgames.Scribble2-v0',
    'flashgames.SeaPong-v0',
    'flashgames.ShamelessClone2-v0',
    'flashgames.Sheepster-v0',
    'flashgames.Sheepy-v0',
    'flashgames.Shift-v0',
    'flashgames.ShimmyChute-v0',
    'flashgames.ShootTheCircle-v0',
    'flashgames.ShortCircuit-v0',
    'flashgames.SiegeHeroPiratePillage-v0',
    'flashgames.Sieger2LevelPack-v0',
    'flashgames.SiegerRebuiltToDestroy-v0',
    'flashgames.Sirtet-v0',
    'flashgames.SistersOfNoMercy-v0',
    'flashgames.SkiSim-v0',
    'flashgames.SkyIsland-v0',
    'flashgames.SkyKnight2-v0',
    'flashgames.SkyQuest-v0',
    'flashgames.Skytrip-v0',
    'flashgames.SliceTheBox-v0',
    'flashgames.SliceTheBoxRemaster-v0',
    'flashgames.SlingBaby-v0',
    'flashgames.SlipSlideSloth-v0',
    'flashgames.SmashTheSwine-v0',
    'flashgames.SmileyJumpFest-v0',
    'flashgames.SmileyPuzzle-v0',
    'flashgames.SmileyPuzzle2-v0',
    'flashgames.SmileyPuzzleGirlEdition-v0',
    'flashgames.SmileyShowdown-v0',
    'flashgames.SnackOnLittleCreatures-v0',
    'flashgames.SnailBob4-v0',
    'flashgames.Snake-v0',
    'flashgames.SnakeClassic-v0',
    'flashgames.SnakeFightArena-v0',
    'flashgames.SneakyScubaEscape-v0',
    'flashgames.SnowPrincessMakeup-v0',
    'flashgames.SnowQueen-v0',
    'flashgames.SnowQueen3-v0',
    'flashgames.SnowQueen4-v0',
    'flashgames.Solarsaurs-v0',
    'flashgames.SonicBubbles-v0',
    'flashgames.SpaceBounty-v0',
    'flashgames.SpaceColony-v0',
    'flashgames.SpaceMadness-v0',
    'flashgames.SpacePunkRacer-v0',
    'flashgames.SpacemanMax-v0',
    'flashgames.SpanishLiga2016-v0',
    'flashgames.Sparks-v0',
    'flashgames.Spectrum-v0',
    'flashgames.SpectrumHeist-v0',
    'flashgames.SpectrumRunner-v0',
    'flashgames.SpeedBusters-v0',
    'flashgames.SpellIdle2-v0',
    'flashgames.SpinClimbGreen-v0',
    'flashgames.SpinSoar-v0',
    'flashgames.SpinSprint-v0',
    'flashgames.SpunkyVsAliens-v0',
    'flashgames.Stalingrad-v0',
    'flashgames.Stalingrad2-v0',
    'flashgames.Stalingrad3-v0',
    'flashgames.Stand-v0',
    'flashgames.StarCars-v0',
    'flashgames.Stardrops-v0',
    'flashgames.Stargrazer-v0',
    'flashgames.Stars-v0',
    'flashgames.Stealthbound-v0',
    'flashgames.StealthboundLevelPack-v0',
    'flashgames.StickBlender-v0',
    'flashgames.StickyNinjaMissions-v0',
    'flashgames.Stickylinky-v0',
    'flashgames.StitchlandConflict-v0',
    'flashgames.StormRage-v0',
    'flashgames.Stratega-v0',
    'flashgames.StreetRace-v0',
    'flashgames.StreetRace2Nitro-v0',
    'flashgames.StreetRace3-v0',
    'flashgames.Streetrace2Nitro-v0',
    'flashgames.StrikeForceKitty-v0',
    'flashgames.SubmarineFighter-v0',
    'flashgames.Sundrops-v0',
    'flashgames.SuperAdventurePalsBattleArena-v0',
    'flashgames.SuperBattleCity2-v0',
    'flashgames.SuperBomb-v0',
    'flashgames.SuperBoxotron2000-v0',
    'flashgames.SuperCandyGems-v0',
    'flashgames.SuperCarRacing-v0',
    'flashgames.SuperDash-v0',
    'flashgames.SuperIdleMaster-v0',
    'flashgames.SuperK9-v0',
    'flashgames.SuperPuzzlePlatformer-v0',
    'flashgames.SuperRally3d-v0',
    'flashgames.SuperRallyChallenge-v0',
    'flashgames.SuperRallyChallenge2-v0',
    'flashgames.SuperRallyExtreme-v0',
    'flashgames.SuperShinyheadHarderThanFlappyBird-v0',
    'flashgames.SuperXtreme5MinuteShootEmUp-v0',
    'flashgames.SuperbikeExtreme-v0',
    'flashgames.SuperbikeRacer-v0',
    'flashgames.SupercarDomination-v0',
    'flashgames.SupergirlGo-v0',
    'flashgames.SurfBuggy-v0',
    'flashgames.SurvivalLab-v0',
    'flashgames.SurvivorMissionD-v0',
    'flashgames.SushiCatTheHoneymoon-v0',
    'flashgames.SwagMan-v0',
    'flashgames.SwampTreck-v0',
    'flashgames.SwapTheDots-v0',
    'flashgames.SweetTooth-v0',
    'flashgames.SwimmingRace-v0',
    'flashgames.SwingTriangle-v0',
    'flashgames.TableTennisChallenge-v0',
    'flashgames.TamusMitta-v0',
    'flashgames.TankStorm-v0',
    'flashgames.TankStorm2-v0',
    'flashgames.TankStorm3-v0',
    'flashgames.TankStorm4-v0',
    'flashgames.TapRocket-v0',
    'flashgames.TastyFruits-v0',
    'flashgames.TattooArtist-v0',
    'flashgames.TaxiInc-v0',
    'flashgames.TaxiRacers-v0',
    'flashgames.TechnoMania-v0',
    'flashgames.TempleRunKnight-v0',
    'flashgames.TerrestrialConflict-v0',
    'flashgames.ThatRedButton-v0',
    'flashgames.Thaw-v0',
    'flashgames.TheBigEscape-v0',
    'flashgames.TheBoomlandsWorldWars-v0',
    'flashgames.TheBravestHunter-v0',
    'flashgames.TheCaseOfScaryShadow-v0',
    'flashgames.TheCubicMonkeyAdventures2-v0',
    'flashgames.TheGreatSiege-v0',
    'flashgames.TheOneForkRestaurantDx-v0',
    'flashgames.ThePretenderPartThree-v0',
    'flashgames.TheProfessionals3-v0',
    'flashgames.TheSilentPlanet-v0',
    'flashgames.TheThreeTowers-v0',
    'flashgames.TheTowerman-v0',
    'flashgames.Thundercars-v0',
    'flashgames.TinyCastle-v0',
    'flashgames.TinyRacers-v0',
    'flashgames.Titanic-v0',
    'flashgames.TokyoGuineaPop-v0',
    'flashgames.ToonEscapeMaze-v0',
    'flashgames.ToonEscapeSpookHouse-v0',
    'flashgames.Tosuta-v0',
    'flashgames.TouchTheBubbles4-v0',
    'flashgames.TouchTheSky-v0',
    'flashgames.TowerCollapseDeluxe-v0',
    'flashgames.TowerEmpire-v0',
    'flashgames.TowerEmpire2-v0',
    'flashgames.TowerMoon-v0',
    'flashgames.TowerOfPisa-v0',
    'flashgames.ToyRacers-v0',
    'flashgames.ToyWarAngryRobotDog-v0',
    'flashgames.TractorTrial-v0',
    'flashgames.TractorTrial2-v0',
    'flashgames.TrafficCollision-v0',
    'flashgames.TrickOrToad-v0',
    'flashgames.TrickyRick-v0',
    'flashgames.Trizzle-v0',
    'flashgames.TrollingLionJump-v0',
    'flashgames.TtRacer-v0',
    'flashgames.TumbleTiles-v0',
    'flashgames.Tumblestump2-v0',
    'flashgames.TurboCrew-v0',
    'flashgames.TurboRally-v0',
    'flashgames.TurtleBreak-v0',
    'flashgames.TutiFruti-v0',
    'flashgames.TwinkleStarRush-v0',
    'flashgames.Typeasaurus-v0',
    'flashgames.UdderChaos-v0',
    'flashgames.UltimateEscape-v0',
    'flashgames.UltimateLegend-v0',
    'flashgames.Underrun-v0',
    'flashgames.UnderwaterSecrets-v0',
    'flashgames.UnfreezeMe3-v0',
    'flashgames.UrbanFatburner-v0',
    'flashgames.UrbanMicroRacers-v0',
    'flashgames.V8MuscleCars-v0',
    'flashgames.V8MuscleCars2-v0',
    'flashgames.V8MuscleCars3-v0',
    'flashgames.V8RacingChampion-v0',
    'flashgames.VanguardWars-v0',
    'flashgames.VectorRunner-v0',
    'flashgames.Velocity-v0',
    'flashgames.VengeanceRider-v0',
    'flashgames.VideoGameMonster-v0',
    'flashgames.ViewtifulFightClub2-v0',
    'flashgames.Viridia-v0',
    'flashgames.VirtualRacer-v0',
    'flashgames.VolcanoPanicInIsland-v0',
    'flashgames.VolleyBomb-v0',
    'flashgames.WackyStrike-v0',
    'flashgames.WarBerlinIdle-v0',
    'flashgames.WarHeroes-v0',
    'flashgames.WarOfTheShard-v0',
    'flashgames.WastelandSiege-v0',
    'flashgames.WaveLucha-v0',
    'flashgames.Weirdville-v0',
    'flashgames.WhatsInsideTheBox-v0',
    'flashgames.Wheelers-v0',
    'flashgames.WhistleAndMice-v0',
    'flashgames.WildWestConflict-v0',
    'flashgames.WilliamTell-v0',
    'flashgames.WindowShooter-v0',
    'flashgames.WinterSlider-v0',
    'flashgames.WishTotems-v0',
    'flashgames.WishTotemsLevelPack-v0',
    'flashgames.WizkidEscape-v0',
    'flashgames.WolfSpiderJigsawPuzzle-v0',
    'flashgames.WonderRocket-v0',
    'flashgames.WoollyBearJigsawPuzzle-v0',
    'flashgames.WorldsGuard2-v0',
    'flashgames.WormHappy-v0',
    'flashgames.WreckRoad-v0',
    'flashgames.XChains-v0',
    'flashgames.XmasChains-v0',
    'flashgames.Xmatch2016-v0',
    'flashgames.Xnake-v0',
    'flashgames.YepisJourney-v0',
    'flashgames.YummyyummyMonsterShooter-v0',
    'flashgames.Zed-v0',
    'flashgames.Zevil2-v0',
    'flashgames.ZodiacMatch-v0',
    'flashgames.Zombality-v0',
    'flashgames.ZombieDemolisher3-v0',
    'flashgames.ZombieMatch3-v0',
    'flashgames.ZombieTdReborn-v0',
    'flashgames.ZombieTowerDefenseReborn-v0',
    'flashgames.ZombiesAndDonuts-v0',
    'flashgames.ZombiesMustDie-v0',
    'flashgames.ZombiesVsBrains-v0',
    'flashgames.Zombonarium-v0',
    'flashgames.ZooRacer-v0',
    ]:
    register(
        id=game,
        entry_point='universe.wrappers:WrappedFlashgamesEnv',
        tags={
            'vnc': True,
            'flashgames': True,
            'runtime': 'flashgames',
            'metadata_encoding': metadata_v1,
            'action_probe': {
                'type': 'key',
                'value': 0x60,
            }
        },
        timestep_limit=20000,
    )

register(
    id='VNCNoopFlashgamesEnv-v0',  # Special noop flashgame env
    entry_point='universe.vnc:WrappedFlashgamesEnv',
    tags={
        'vnc': True,
        'flashgames': True,
        'runtime': 'flashgames',
    },
    timestep_limit=10**7,
)

#------------------------ World of Bits and MiniWoB ------------------------#
#     "World of Bits" comprises a series of browser tasks,
#     including a series of simple "MiniWoB" tasks such
#     as using buttons and sliders, as well as more complex
#     tasks such as booking flights on actual websites.

vnc_world_of_bits = [
    'wob.MiniWorldOfBits-v0',
    'wob.mini.BisectAngle-v0',
    'wob.mini.BookFlight-v0',
    'wob.mini.ChaseCircle-v0',
    'wob.mini.ChooseDate-v0',
    'wob.mini.ChooseList-v0',
    'wob.mini.CircleCenter-v0',
    'wob.mini.ClickButton-v0',
    'wob.mini.ClickButtonSequence-v0',
    'wob.mini.ClickCheckboxes-v0',
    'wob.mini.ClickCollapsible-v0',
    'wob.mini.ClickCollapsible2-v0',
    'wob.mini.ClickColor-v0',
    'wob.mini.ClickDialog-v0',
    'wob.mini.ClickDialog2-v0',
    'wob.mini.ClickLink-v0',
    'wob.mini.ClickMenu-v0',
    'wob.mini.ClickMenu2-v0',
    'wob.mini.ClickOption-v0',
    'wob.mini.ClickPie-v0',
    'wob.mini.ClickScrollList-v0',
    'wob.mini.ClickShades-v0',
    'wob.mini.ClickShape-v0',
    'wob.mini.ClickTab-v0',
    'wob.mini.ClickTab2-v0',
    'wob.mini.ClickTest-v0',
    'wob.mini.ClickTest2-v0',
    'wob.mini.ClickWidget-v0',
    'wob.mini.CopyPaste-v0',
    'wob.mini.CopyPaste2-v0',
    'wob.mini.CountShape-v0',
    'wob.mini.CountSides-v0',
    'wob.mini.DragBox-v0',
    'wob.mini.DragCube-v0',
    'wob.mini.DragItem-v0',
    'wob.mini.DragItems-v0',
    'wob.mini.DragItemsGrid-v0',
    'wob.mini.DragShapes-v0',
    'wob.mini.DragSortNumbers-v0',
    'wob.mini.EmailInbox-v0',
    'wob.mini.EnterDate-v0',
    'wob.mini.EnterPassword-v0',
    'wob.mini.EnterText-v0',
    'wob.mini.EnterText2-v0',
    'wob.mini.EnterTextDynamic-v0',
    'wob.mini.EnterTime-v0',
    'wob.mini.FindMidpoint-v0',
    'wob.mini.FindWord-v0',
    'wob.mini.FocusText-v0',
    'wob.mini.FocusText2-v0',
    'wob.mini.GridCoordinate-v0',
    'wob.mini.GuessNumber-v0',
    'wob.mini.HighlightText-v0',
    'wob.mini.HighlightText2-v0',
    'wob.mini.IdentifyShape-v0',
    'wob.mini.LoginUser-v0',
    'wob.mini.MovingItems-v0',
    'wob.mini.NavigateTree-v0',
    'wob.mini.NumberCheckboxes-v0',
    'wob.mini.ReadTable-v0',
    'wob.mini.ReadTable2-v0',
    'wob.mini.ResizeTextarea-v0',
    'wob.mini.RightAngle-v0',
    'wob.mini.ScrollText-v0',
    'wob.mini.ScrollText2-v0',
    'wob.mini.SearchEngine-v0',
    'wob.mini.SimonSays-v0',
    'wob.mini.SimpleAlgebra-v0',
    'wob.mini.SimpleArithmetic-v0',
    'wob.mini.SocialMedia-v0',
    'wob.mini.Terminal-v0',
    'wob.mini.TextEditor-v0',
    'wob.mini.TextTransform-v0',
    'wob.mini.TicTacToe-v0',
    'wob.mini.UseAutocomplete-v0',
    'wob.mini.UseColorwheel-v0',
    'wob.mini.UseColorwheel2-v0',
    'wob.mini.UseSlider-v0',
    'wob.mini.UseSlider2-v0',
    'wob.mini.UseSpinner-v0',
    'wob.mini.VisualAddition-v0',
]

# signup forms.
for _id in range(20):
    vnc_world_of_bits.append('wob.real.Signup-{}-v0'.format(_id))

for _site in ['Jetblue', 'Kayak', 'AA', 'VirginAmerica',
               'United',  'Delta',  'Alaska']:
    vnc_world_of_bits.append('wob.real.BookFlight-{}-v0'.format(_site))

for _site in ['Airfrance', 'Craigslist', 'Chase']:
    vnc_world_of_bits.append('wob.real.ClickButton-{}-v0'.format(_site))

for _task in ['Learn', 'Test']:
    for _name in ['Geography', 'Planet', 'Universe', 'Comet', 'Moon', 'Mars', 'Solar-System']:
        vnc_world_of_bits.append('wob.real.Quizlet-{}-{}-v0'.format(_name, _task))

vnc_world_of_bits.append('wob.real.Duolingo-French-Basic-1-v0')

for game in vnc_world_of_bits:
    register(
        id=game,
        entry_point='universe.wrappers:WrappedVNCEnv',
        tags={
            'vnc': True,
            'wob': True,
            'runtime': 'world-of-bits'
        },
        timestep_limit=10**7,
    )

#-------------------------- Complex Games ------------------------#
#     Any game, program, app, or website can be a
#     Universe environment. Here we include
#     a handful of sample "complex" games
#     such as World of Bits, GTA V, and StarCraft.
#     Adding more games is straightforward, and
#     we welcome contributions of environments
#     from the community!

for id in ['starcraft.TerranAstralBalance-v0']:
    register(
        id=id,
        entry_point='universe.wrappers:WrappedStarCraftEnv',
        tags={
            'vnc': True,
            'starcraft': True,
            'runtime': 'starcraft',
        },
        timestep_limit=10**7,
    )

for gtav_game in ['gtav.SaneDriving-v0', 'gtav.Speed-v0']:
    register(
        id=gtav_game,
        entry_point='universe.wrappers:WrappedGTAVEnv',
        tags={
            'vnc': True,
            'gtav': True,
            'runtime': 'vnc-windows',
        },
        timestep_limit=10**7,
    )

register(
    id='world.WorldOfGoo-v0',
    entry_point='universe.wrappers:WrappedWorldOfGooEnv',
    tags={
        'vnc': True,
        'wog': True,
        'runtime': 'vnc-world-of-goo',
    },
    timestep_limit=10**7,
)

for slith_game in ['SlitherIO-v0', 'SlitherIONoSkins-v0', 'SlitherIOEasy-v0']:
    register(
        id='internet.' + slith_game,
        entry_point='universe.wrappers:WrappedInternetEnv',
        tags={
            'vnc': True,
            'internet': True,
            'slither': True,
            'runtime': 'flashgames',
            'metadata_encoding': metadata_v1,
            'action_probe': {
                'type': 'key',
                'value': 0x60,
            }
        },
        timestep_limit=10**7,
    )

register(
    id='test.DummyVNCEnv-v0',
    entry_point='universe.envs:DummyVNCEnv',
    tags={
        'vnc': True,
        'metadata_encoding': metadata_v1,
        'action_probe': {
            'type': 'key',
            'value': 0x60,
        }
    },
    timestep_limit=10**7
    )



import re
from universe import error
from universe.remotes.allocator_remote import AllocatorManager
from universe.remotes.docker_remote import DockerManager
from universe.remotes.hardcoded_addresses import HardcodedAddresses

def build(client_id, remotes, runtime=None, start_timeout=None, **kwargs):
    if isinstance(remotes, int):
        remotes = str(remotes)
    elif not isinstance(remotes, str):
        raise error.Error('remotes argument must be a string, got {} which is of type {}'.format(remotes, type(remotes)))

    if re.search('^\d+$', remotes): # an integer, like -r 20
        n = int(remotes)
        return DockerManager(
            runtime=runtime,
            start_timeout=start_timeout,
            reuse=kwargs.get('reuse', False),
            n=n,
        ), n
    elif remotes.startswith('vnc://'):
        return HardcodedAddresses.build(
            remotes,
            start_timeout=start_timeout)
    elif remotes.startswith('http://') or remotes.startswith('https://'):
        if runtime is None:
            raise error.Error('Must provide a runtime. HINT: try creating your env instance via gym.make("flashgames.DuskDrive-v0")')

        manager, n = AllocatorManager.from_remotes(
            client_id,
            remotes,
            runtime_id=runtime.id,
            tag=runtime.image.split(':')[-1],
            start_timeout=start_timeout,
            api_key=kwargs.get('api_key'),
            use_recorder_ports=kwargs.get('use_recorder_ports', False),
        )
        manager.start()
        return manager, n
    else:
        raise error.Error('Invalid remotes: {!r}. Must be an integer or must start with vnc:// or https://'.format(remotes))



from autobahn.twisted import websocket
import logging
import numpy as np
import threading
import time

from twisted.python import failure
from twisted.internet import defer, endpoints
import twisted.internet.error

from universe import utils
from universe.twisty import reactor
from universe.rewarder import connection_timer, env_status, reward_buffer, rewarder_client
from universe.utils import display

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

def _ping(client):
    return client.send('v0.control.ping', {}, expect_reply=True)

class RewarderSession(object):
    def __init__(self):
        self.lock = threading.RLock()

        self.i = 0

        # Mutated by main thread exclusively
        self.names_by_id = {}
        self.reward_buffers = {}
        self.env_statuses = {}
        self.errors = {}
        self.networks = {}

        self.clients = {}

    def close(self, name=None, reason=u'closed by RewarderSession.close'):
        if name is None:
            names = list(self.names_by_id.values())
        else:
            logger.info('[%s] Closing rewarder connection', name)
            names = [name]
        self.ids_by_name = {name: id for id, name in self.names_by_id.items()}

        for name in names:
            with self.lock:
                id = self.ids_by_name.pop(name, None)
                if id is None:
                    # already closed
                    continue

                del self.names_by_id[id]
                del self.reward_buffers[id]
                del self.env_statuses[id]
                self.errors.pop(id, None)

                network = self.networks.pop(id)
                network.close()

                client = self.clients.pop(id, None)
                if client is not None:
                    reactor.callFromThread(client.close, reason=reason)

    def connect(self, name, address, label, password, env_id=None, seed=None, fps=60,
                start_timeout=None, observer=False, skip_network_calibration=False):
        if name in self.reward_buffers:
            self.close(name, reason='closing previous connection to reconnect with the same name')

        network = Network()
        self.names_by_id[self.i] = name
        self.reward_buffers[self.i] = reward_buffer.RewardBuffer(label)
        self.env_statuses[self.i] = env_status.EnvStatus(label=label, primary=False)
        self.networks[self.i] = network

        reactor.callFromThread(self._connect,
                               name=name,
                               address=address,
                               env_id=env_id,
                               seed=seed,
                               fps=fps,
                               i=self.i,
                               network=network,
                               env_status=self.env_statuses[self.i],
                               reward_buffer=self.reward_buffers[self.i],
                               label=label,
                               start_timeout=start_timeout,
                               password=password,
                               observer=observer,
                               skip_network_calibration=skip_network_calibration,
        )
        self.i += 1
        return network

    def _already_closed(self, i):
        # Lock must be held
        return i not in self.names_by_id

    # Call only from Twisted thread

    # TODO: probably time to convert to kwargs
    @defer.inlineCallbacks
    def _connect(self, name, address, env_id, seed, fps, i, network, env_status, reward_buffer,
                 label, password, start_timeout,
                 observer, skip_network_calibration,
                 attempt=0, elapsed_sleep_time=0,
    ):
        endpoint = endpoints.clientFromString(reactor, 'tcp:'+address)
        factory = websocket.WebSocketClientFactory('ws://'+address)
        factory.protocol = rewarder_client.RewarderClient

        assert password, "Missing password: {} for rewarder session".format(password)
        factory.headers = {'authorization': utils.basic_auth_encode(password), 'openai-observer': 'true' if observer else 'false'}
        factory.i = i

        # Various important objects
        factory.endpoint = endpoint
        factory.env_status = env_status
        factory.reward_buffer = reward_buffer

        # Helpful strings
        factory.label = label
        factory.address = address

        # Arguments to always send to the remote reset call
        factory.arg_env_id = env_id
        factory.arg_fps = fps

        def record_error(e):
            if isinstance(e, failure.Failure):
                e = e.value

            # logger.error('[%s] Recording rewarder error: %s', factory.label, e)
            with self.lock:
                # drop error on the floor if we're already closed
                if self._already_closed(factory.i):
                    extra_logger.info('[%s] Ignoring error for already closed connection: %s', label, e)
                elif factory.i not in self.clients:
                    extra_logger.info('[%s] Received error for connection which has not been fully initialized: %s', label, e)
                    # We could handle this better, but right now we
                    # just mark this as a fatal error for the
                    # backend. Often it actually is.
                    self.errors[factory.i] = e
                else:
                    extra_logger.info('[%s] Recording fatal error for connection: %s', label, e)
                    self.errors[factory.i] = e

        def retriable_error(e, error_message):
            if isinstance(e, failure.Failure):
                e = e.value

            if self._already_closed(factory.i):
                logger.error('[%s] Got error, but giving up on reconnecting, since %d already disconnected', factory.label, factory.i)
                return

            # Also need to handle DNS errors, so let's just handle everything for now.
            #
            # reason.trap(twisted.internet.error.ConnectError, error.ConnectionError)
            if elapsed_sleep_time < start_timeout:
                sleep = min((2 * attempt+1), 10)
                logger.error('[%s] Waiting on rewarder: %s. Retry in %ds (slept %ds/%ds): %s', factory.label, error_message, sleep, elapsed_sleep_time, start_timeout, e)
                reactor.callLater(
                    sleep, self._connect, name=name, address=address,
                    env_id=env_id, seed=seed, fps=fps, i=i, network=network,
                    env_status=env_status, reward_buffer=reward_buffer, label=label,
                    attempt=attempt+1, elapsed_sleep_time=elapsed_sleep_time+sleep,
                    start_timeout=start_timeout, password=password,
                    observer=observer, skip_network_calibration=skip_network_calibration,
                )
            else:
                logger.error('[%s] %s. Retries exceeded (slept %ds/%ds): %s', factory.label, error_message, elapsed_sleep_time, start_timeout, e)
                record_error(e)

        factory.record_error = record_error

        try:
            retry_msg = 'establish rewarder TCP connection'
            client = yield endpoint.connect(factory)
            extra_logger.info('[%s] Rewarder TCP connection established', factory.label)

            retry_msg = 'complete WebSocket handshake'
            yield client.waitForWebsocketConnection()
            extra_logger.info('[%s] Websocket client successfully connected', factory.label)

            if not skip_network_calibration:
                retry_msg = 'run network calibration'
                yield network.calibrate(client)
                extra_logger.info('[%s] Network calibration complete', factory.label)

            retry_msg = ''

            if factory.arg_env_id is not None:
                # We aren't picky about episode ID: we may have
                # already receieved an env.describe message
                # telling us about a resetting environment, which
                # we don't need to bump post.
                #
                # tl;dr hardcoding 0.0 here avoids a double reset.
                reply = yield self._send_env_reset(client, seed=seed, episode_id='0')
            else:
                # No env_id requested, so we just proceed without a reset
                reply = None
            # We're connected and have measured the
            # network. Mark everything as ready to go.
            with self.lock:
                if factory.i not in self.names_by_id:
                    # ID has been popped!
                    logger.info('[%s] Rewarder %d started, but has already been closed', factory.label, factory.i)
                    client.close(reason='RewarderSession: double-closing, client was closed while RewarderSession was starting')
                elif reply is None:
                    logger.info('[%s] Attached to running environment without reset', factory.label)
                else:
                    context, req, rep = reply
                    logger.info('[%s] Initial reset complete: episode_id=%s', factory.label, rep['headers']['episode_id'])
                self.clients[factory.i] = client
        except Exception as e:
            if retry_msg:
                retriable_error(e, 'failed to ' + retry_msg)
            else:
                record_error(e)

    def pop_errors(self):
        errors = {}
        with self.lock:
            if self.errors:
                for i, error in self.errors.items():
                    name = self.names_by_id[i]
                    errors[name] = error
                self.errors.clear()
        return errors

    def reset(self, seed=None):
        with self.lock:
            for i, reward_buffer in self.reward_buffers.items():
                reward_buffer.mask()
        reactor.callFromThread(self._reset, seed=seed)

    def _reset(self, seed=None):
        with self.lock:
            for client in self.clients.values():
                d = self._send_env_reset(client, seed=seed)
                # Total hack to capture the variable in the closure
                def callbacks(client):
                    def success(reply): pass
                    def fail(reason): client.factory.record_error(reason)
                    return success, fail
                success, fail = callbacks(client)
                d.addCallback(success)
                d.addErrback(fail)

    def _send_env_reset(self, client, seed=None, episode_id=None):
        if episode_id is None:
            episode_id = client.factory.env_status.episode_id
        logger.info('[%s] Sending reset for env_id=%s fps=%s episode_id=%s', client.factory.label, client.factory.arg_env_id, client.factory.arg_fps, episode_id)
        return client.send_reset(
            env_id=client.factory.arg_env_id,
            seed=seed,
            fps=client.factory.arg_fps,
            episode_id=episode_id)

    def pop(self, warn=True, peek_d=None):
        reward_d = {}
        done_d = {}
        info_d = {}
        err_d = self.pop_errors()

        for i, reward_buffer in self.reward_buffers.items():
            name = self.names_by_id[i]

            reward, done, info = reward_buffer.pop(peek_d.get(name))
            reward_d[name] = reward
            done_d[name] = done
            info_d[name] = info

        # TODO: use FPS here rather than 60
        if warn and any(info.get('stats.reward.count', 0) > 60 for info in info_d.values()):
            logger.warn('WARNING: returning more than 60 aggregated rewards: %s. Either your agent is not keeping up with the framerate, or you should have called ".reset()" to clear pending rewards and reset the environments to a known state.',
                        {name: '{} (episode_id={})'.format(info['stats.reward.count'], info.get('env_status.episode_id')) for name, info in info_d.items()})

        return reward_d, done_d, info_d, err_d

    def wait(self, timeout=None):
        deadline = time.time() + timeout
        for client in self.clients:
            if timeout is not None:
                remaining_timeout = deadline - time.time()
            else:
                remaining_timeout = None
            client.reward_buffer.wait_for_step(timeout=remaining_timeout)

    # Hack to test actions over websockets
    # TODO: Carve websockets out of rewarder pkg (into vnc_env? - and move this there)
    def send_action(self, action_n, env_id):
        reactor.callFromThread(self._send_action, env_id, action_n)
        return self.pop_errors()

    def _send_action(self, env_id, action_n):
        with self.lock:
            for n, client in zip(action_n, self.clients.values()):
                self._send_env_action(client, env_id, action_n[n])

    def _send_env_action(self, client, env_id, action_n):
        if len(action_n) == 0:
            # Hack to skip empty actions. TODO: Find source (throttle?) and fix
            return
        message = {
            'env_id': env_id,
            'action': action_n,
        }
        client.send('v0.agent.action', message, expect_reply=False)

    def rewards_count(self):
        # TODO: any reason to lock these?
        return [client.reward_buffer.count for client in self.clients]

    def pop_observation(self):
        return [client.reward_buffer.pop_observation() for client in self.clients]

    # def _connection_time(self):
    #     deferreds = []
    #     for client in self.clients:
    #         endpoint = client.factory.endpoint
    #         d = connection_timer.start(endpoint)
    #         deferreds.append(d)

    #     d = defer.DeferredList(deferreds, fireOnOneErrback=True, consumeErrors=True)
    #     return d

# Run this in Twisty therad
class Network(object):
    def __init__(self):
        self.connection_samples = 10
        self.application_ping_samples = 10

        self.connection_time_m = None
        self.lock = threading.Lock()

        self.recalibrate = None
        self.client = None

        self._ntpdate_reversed_clock_skew = None
        self._reversed_clock_skew = None

    def active(self):
        with self.lock:
            return self._reversed_clock_skew is not None

    # Used by external consumers
    def reversed_clock_skew(self):
        with self.lock:
            if self._ntpdate_clock_skew is not None:
                return self._ntpdate_reversed_clock_skew
            else:
                return self._reversed_clock_skew

    def _report(self):
        connection_time = display.display_timestamps(self.connection_time_m)
        if self._ntpdate_clock_skew is not None:
            ntpdate_clock_skew = display.display_timestamp(self._ntpdate_clock_skew[0])
        else:
            ntpdate_clock_skew = None
        clock_skew = display.display_timestamps_pair(self.clock_skew_m)
        application_rtt = display.display_timestamps(self.application_rtt_m)
        request_overhead = display.display_timestamps(self.request_overhead_m)
        response_overhead = display.display_timestamps(self.response_overhead_m)

        extra_logger.info('[%s] Network calibration: ntpdate_clock_skew=%s clock_skew=%s connection_time=%s application_rtt=%s request_overhead=%s response_overhead=%s',
                    self.client.factory.label, ntpdate_clock_skew, clock_skew, connection_time, application_rtt,
                    request_overhead, response_overhead)

    def _start(self):
        def calibrate():
            d = defer.Deferred()
            def fail(reason):
                logger.error('[%s] Could not recalibrate network: %s', self.client.factory.label, reason)
            d.addErrback(fail)
            self._start_measure_connection_time(d)
            self._start()
        self.recalibrate = reactor.callLater(5 * 60, calibrate)

    def close(self):
        if self.recalibrate:
            try:
                self.recalibrate.cancel()
            except twisted.internet.error.AlreadyCalled:
                pass

    # Called externally
    def calibrate(self, client):
        d = defer.Deferred()
        def success(res):
            # If we succeed, kick off the periodic 5 minute
            # recalibrations.
            self._start()
            return res
        d.addCallback(success)

        self.client = client

        # Kinda a hack. Idea is to try using the ntpdate -q offset if
        # we can.
        skew = self._start_measure_clock_skew()
        def succeed(offset):
            with self.lock:
                self._ntpdate_clock_skew = np.array([offset, offset])
                self._ntpdate_reversed_clock_skew = np.array([-offset, -offset])
            self._start_measure_connection_time(d)
        skew.addCallback(succeed)

        def fail(reason):
            with self.lock:
                self._ntpdate_clock_skew = None
                self._ntpdate_reversed_clock_skew = None

            extra_logger.info('[%s] Could not determine clock skew with ntpdate; falling back to application-level ping: %s', self.client.factory.label, reason.value)
            self._start_measure_connection_time(d)
        skew.addErrback(fail)

        return d

    def _start_measure_connection_time(self, d):
        connection_time_m = np.zeros(self.connection_samples)
        self._measure_connection_time(d, connection_time_m, 0)

    def _measure_connection_time(self, d, connection_time_m, i):
        extra_logger.debug('[%s] Measuring connection time (%d/%d)', self.client.factory.label, i+1, len(connection_time_m))
        endpoint = self.client.factory.endpoint
        timer = connection_timer.start(endpoint)

        def success(delta):
            connection_time_m[i] = delta
            if i+1 < len(connection_time_m):
                self._measure_connection_time(d, connection_time_m, i+1)
            else:
                self.connection_time_m = connection_time_m
                self._start_measure_application_ping(d)
        def fail(reason):
            d.errback(reason)
        timer.addCallback(success)
        timer.addErrback(fail)

    def _start_measure_application_ping(self, d=None):
        clock_skew_m = np.zeros((self.application_ping_samples, 2))
        request_overhead_m = np.zeros((self.application_ping_samples))
        response_overhead_m = np.zeros((self.application_ping_samples))
        application_rtt_m = np.zeros((self.application_ping_samples))

        self._measure_application_ping(d, clock_skew_m, request_overhead_m, response_overhead_m, application_rtt_m, 0)

    def _measure_application_ping(self, d, clock_skew_m, request_overhead_m, response_overhead_m, application_rtt_m, i):
        extra_logger.debug('[%s] Issuing an application-level ping (%d/%d)', self.client.factory.label, i+1, len(clock_skew_m))
        start = time.time()
        ping = _ping(self.client)

        def success(res):
            context, request, response = res
            end = time.time()

            request_sent_at = request['headers']['sent_at'] # local
            response_sent_at = response['headers']['sent_at'] # remote
            response_received_at = context['start'] # local

            # We try to put bounds on clock skew by subtracting
            # local and remote times, for local and remote events
            # that are causally related.
            #
            # For example, suppose that the following local/remote
            # logical timestamps apply to a request (for a system
            # with clock skew of 100):
            #
            # request_sent       local: 0   remote: 100
            # request_recieved   local: 1   remote: 101
            # response_sent      local: 2   remote: 102
            # response_received  local: 3   remote: 103
            #
            # Then:
            #
            # # Remote event *after* local is upper bound
            # request_recieved.remote - request_sent.local = 101
            # # Remote event *before* local is lower bound
            # response_sent.remote - response_received.local = 102 - 3 = 99
            #
            # There's danger of further clock drift over time, but
            # we don't need these to be fully accurate, and this
            # should be fine for now.
            clock_skew_m[i, :] = (response_sent_at-response_received_at, response_sent_at-request_sent_at)
            request_overhead_m[i] = request_sent_at - start
            response_overhead_m[i] = end - response_received_at
            application_rtt_m[i] = response_received_at - request_sent_at

            if i+1 < len(clock_skew_m):
                self._measure_application_ping(d, clock_skew_m, request_overhead_m, response_overhead_m, application_rtt_m, i+1)
            else:
                self.clock_skew_m = clock_skew_m
                self.request_overhead_m = request_overhead_m
                self.response_overhead_m = response_overhead_m
                self.application_rtt_m = application_rtt_m

                self._report()
                self._update_exposed_metrics()

                # Ok, all done!
                if d is not None:
                    d.callback(self)
        ping.addCallback(success)
        ping.addErrback(d.errback)

    def _update_exposed_metrics(self):
        with self.lock:
            self._clock_skew = self.clock_skew_m.mean(axis=0) # add to local time to get remote time, as (min, max) values
            self._reversed_clock_skew = -self._clock_skew[[1, 0]] # add to remote time to get local time, in format (min, max)


    def _start_measure_clock_skew(self):
        host = self.client.factory.address.split(':')[0]
        return connection_timer.measure_clock_skew(self.client.factory.label, host)



# a proxy server that handles both reward channel and vnc.
from twisted.python import log
from autobahn.twisted import websocket
import logging
import os
import time
import pexpect
import sys
import threading

from universe.vncdriver.vnc_proxy_server import VNCProxyServer
from universe.rewarder.reward_proxy_server import RewardProxyServer
from universe import utils

logger = logging.getLogger(__name__)


class DualProxyServer(VNCProxyServer):
    def __init__(self, action_queue=None, error_buffer=None, enable_logging=True):
        self._log_info('DualProxyServer inited')
        self.reward_proxy = None

        super(DualProxyServer, self).__init__(action_queue, error_buffer, enable_logging)

    def _log_info(self, msg, *args, **kwargs):
        logger.info('[dual_proxy] ' + msg, *args, **kwargs)

    def recv_ClientInit(self, block):
        # start reward proxy.
        self._log_info('Starting reward proxy server')
        self.reward_proxy = pexpect.spawnu(self.factory.reward_proxy_bin,
                                           logfile=sys.stdout,
                                           timeout=None)

        # wait on reward proxy to be up.
        self._log_info('Waiting for reward proxy server')
        self.reward_proxy.expect('\[RewardProxyServer\]')
        self.reward_proxy_thread = threading.Thread(target=lambda: self.reward_proxy.expect(pexpect.EOF))
        self.reward_proxy_thread.start()

        self._log_info('Reward proxy server is up %s', self.reward_proxy.before)

        super(DualProxyServer, self).recv_ClientInit(block)

        self.logfile_dir = self.log_manager.logfile_dir

    def close(self):
        # end connections.
        super(DualProxyServer, self).close()

        # wait for rewarder to close.
        if self.reward_proxy:
            self.reward_proxy.terminate()

        # upload to s3.
        # probably hacky right now.
        logger.info('log manager = %s', self.log_manager)
        if self.log_manager:
            os.system('/app/universe/bin/upload_directory.sh demonstrator_%(recorder_id)s %(directory)s %(bucket)s' %
                    dict(recorder_id=self.factory.recorder_id, directory=self.logfile_dir,
                        bucket=self.factory.bucket)
                    )




import gym
import weakref

from universe import error
from universe.vectorized import core

class Vectorize(gym.Wrapper):
    """
Given an unvectorized environment (where, e.g., the output of .step() is an observation
rather than a list of observations), turn it into a vectorized environment with a batch of size
1.
"""

    metadata = {'runtime.vectorized': True}

    def __init__(self, env):
        super(Vectorize, self).__init__(env)
        assert not env.metadata.get('runtime.vectorized')
        assert self.metadata.get('runtime.vectorized')
        self.n = 1

    def _reset(self):
        observation = self.env.reset()
        return [observation]

    def _step(self, action):
        observation, reward, done, info = self.env.step(action[0])
        return [observation], [reward], [done], {'n': [info]}

    def _seed(self, seed):
        return [self.env.seed(seed[0])]

class Unvectorize(core.Wrapper):
    """
Take a vectorized environment with a batch of size 1 and turn it into an unvectorized environment.
""" 
    autovectorize = False
    metadata = {'runtime.vectorized': False}

    def _configure(self, **kwargs):
        super(Unvectorize, self)._configure(**kwargs)
        if self.n != 1:
            raise error.Error('Can only disable vectorization with n=1, not n={}'.format(self.n))

    def _reset(self):
        observation_n = self.env.reset()
        return observation_n[0]

    def _step(self, action):
        action_n = [action]
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        return observation_n[0], reward_n[0], done_n[0], info['n'][0]

    def _seed(self, seed):
        return self.env.seed([seed])[0]

class WeakUnvectorize(Unvectorize):
    def __init__(self, env, i):
        self._env_ref = weakref.ref(env)
        super(WeakUnvectorize, self).__init__(env)
        # WeakUnvectorize won't get configure called on it
        self.i = i

    def _check_for_duplicate_wrappers(self):
        pass  # Disable this check because we need to wrap vectorized envs in multiple unvectorize wrappers

    @property
    def env(self):
        # Called upon instantiation
        if not hasattr(self, '_env_ref'):
            return

        env = self._env_ref()
        if env is None:
            raise error.Error("env has been garbage collected. To keep using WeakUnvectorize, you must keep around a reference to the env object. (HINT: try assigning the env to a variable in your code.)")
        return env

    @env.setter
    def env(self, value):
        # We'll maintain our own weakref, thank you very much.
        pass

    def _seed(self, seed):
        # We handle the seeding ourselves in the vectorized Monitor
        return [seed]

    def close(self):
        # Don't want to close through this wrapper
        pass



try:
    # In Py2, use the more efficient cStringIO implementation if it's
    # available
    from cStringIO import StringIO as BytesIO
except ImportError:
    # Fall back to using normal BytesIO, six handles python 2 vs 3 compat
    from six import BytesIO

import logging
import numpy as np
from universe import pyprofile
import struct

logger = logging.getLogger(__name__)

class FramebufferUpdate(object):
    def __init__(self, rectangles):
        self.rectangles = rectangles

class Rectangle(object):
    def __init__(self, x, y, width, height, encoding):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.encoding = encoding

class PseudoCursorEncoding(object):
    def __init__(self, image, mask):
        self.image = image
        self.mask = mask

    @classmethod
    def parse_rectangle(cls, client, x, y, width, height, data):
        split = width * height * client.framebuffer.bypp
        image = np.frombuffer(data[:split], np.uint8).reshape((height, width, 4))[:, :, [0, 1, 2]]

        # Turn raw bytes into uint8 array
        mask = np.frombuffer(data[split:], np.uint8)
        # Turn uint8 array into bit array, and go over the scanlines
        mask = np.unpackbits(mask).reshape((height, -1 if mask.size else 0))[:, :width]

        encoding = cls(image, mask)
        return Rectangle(x, y, width, height, encoding)

class RAWEncoding(object):
    def __init__(self, data):
        self.data = data

    @classmethod
    def parse_rectangle(cls, client, x, y, width, height, data):
        assert client.framebuffer.bpp == 32
        data = np.frombuffer(data, np.uint8).reshape((height, width, 4))[:, :, [0, 1, 2]]
        encoding = cls(data)
        return Rectangle(x, y, width, height, encoding)

class ZlibEncoding(object):
    def __init__(self, data):
        self.data = data

    @classmethod
    def parse_rectangle(cls, client, x, y, width, height, data):
        decompressed = client.zlib_decompressor.decompress(data)
        logger.debug('[zlib] Decompressed from %s bytes -> %s bytes', len(data), len(decompressed))
        pyprofile.incr('vncdriver.recv_rectangle.zlib_encoding.decompressed_bytes', len(decompressed), unit=pyprofile.BYTES)
        data = np.frombuffer(decompressed, np.uint8).reshape((height, width, 4))[:, :, [0, 1, 2]]
        encoding = cls(data)
        return Rectangle(x, y, width, height, encoding)

class ZRLEEncoding(object):
    def __init__(self, data):
        self.data = data

    @classmethod
    def parse_rectangle(cls, client, x, y, width, height, data):
        decompressed = client.zlib_decompressor.decompress(data)
        logger.debug('[zrle] Decompressed from %s bytes -> %s bytes', len(data), len(decompressed))
        pyprofile.incr('vncdriver.recv_rectangle.zrle_encoding.decompressed_bytes', len(decompressed), unit=pyprofile.BYTES)

        if client.framebuffer.bpp > 24:
            bytes_per_pixel = 3
        else:
            bytes_per_pixel = client.framebuffer.bypp

        buf = BytesIO(decompressed)
        data = cls._read(x, y, width, height, buf, bytes_per_pixel)
        encoding = cls(data)
        return Rectangle(x, y, width, height, encoding)

    @classmethod
    def _read(cls, x, y, width, height, buf, bytes_per_pixel):
        data = np.zeros([height, width, 3], dtype=np.uint8) + 255

        for tile_y in range(0, height, 64):
            tile_height = min(64, height-tile_y)

            for tile_x in range(0, width, 64):
                tile_width = min(64, width-tile_x)

                tile = data[tile_y:tile_y+tile_height, tile_x:tile_x+tile_width]
                cls._read_tile(tile, buf, tile_width, tile_height, bytes_per_pixel)
        return data

    @classmethod
    def _read_tile(cls, tile, buf, tile_width, tile_height, bytes_per_pixel):
        assert bytes_per_pixel == 3

        # Each tile begins with a subencoding type byte.  The top bit of this
        # byte is set if the tile has been run-length encoded, clear otherwise.
        # The bottom 7 bits indicate the size of the palette used: zero means
        # no palette, 1 means that the tile is of a single color, and 2 to 127
        # indicate a palette of that size.  The special subencoding values 129
        # and 127 indicate that the palette is to be reused from the last tile
        # that had a palette, with and without RLE, respectively.
        (subencoding,) = struct.unpack('!B', buf.read(1))

        run_length_encoded = bool(subencoding & 128)
        palette_size = subencoding & 127

        bytes = palette_size * bytes_per_pixel
        palette_data = buf.read(bytes)
        assert len(palette_data) == bytes, "Palette data came up short: {} bytes rather than {}".format(len(palette_data), bytes)

        logger.debug('Handling zrle tile: run_length_encoded=%s palette_size=%s', run_length_encoded, palette_size)

        if palette_size == 0 and not run_length_encoded:
            # 0: Raw pixel data. width*height pixel values follow (where width and
            # height are the width and height of the tile):
            #
            #  +-----------------------------+--------------+-------------+
            #  | No. of bytes                | Type [Value] | Description |
            #  +-----------------------------+--------------+-------------+
            #  | width*height*BytesPerCPixel | CPIXEL array | pixels      |
            #  +-----------------------------+--------------+-------------+
            data = buf.read(bytes_per_pixel * tile_width * tile_height)
            data = np.frombuffer(data, dtype=np.uint8).reshape((tile_height, tile_width, 3))
            tile[:, :, :] = data
            return
        elif palette_size == 1 and not run_length_encoded:
            # 1: A solid tile consisting of a single color.  The pixel value
            # follows:
            #
            # +----------------+--------------+-------------+
            # | No. of bytes   | Type [Value] | Description |
            # +----------------+--------------+-------------+
            # | bytesPerCPixel | CPIXEL       | pixelValue  |
            # +----------------+--------------+-------------+
            color = np.frombuffer(palette_data, dtype=np.uint8).reshape((3, ))
            tile[:, :, :] = color
            return
        elif not run_length_encoded:
            # 2 to 16:  Packed palette types.  The paletteSize is the value of the
            # subencoding, which is followed by the palette, consisting of
            # paletteSize pixel values.  The packed pixels follow, with each
            # pixel represented as a bit field yielding a zero-based index into
            # the palette.  For paletteSize 2, a 1-bit field is used; for
            # paletteSize 3 or 4, a 2-bit field is used; and for paletteSize
            # from 5 to 16, a 4-bit field is used.  The bit fields are packed
            # into bytes, with the most significant bits representing the
            # leftmost pixel (i.e., big endian).  For tiles not a multiple of 8,
            # 4, or 2 pixels wide (as appropriate), padding bits are used to
            # align each row to an exact number of bytes.

            #   +----------------------------+--------------+--------------+
            #   | No. of bytes               | Type [Value] | Description  |
            #   +----------------------------+--------------+--------------+
            #   | paletteSize*bytesPerCPixel | CPIXEL array | palette      |
            #   | m                          | U8 array     | packedPixels |
            #   +----------------------------+--------------+--------------+

            #  where m is the number of bytes representing the packed pixels.
            #  For paletteSize of 2, this is div(width+7,8)*height; for
            #  paletteSize of 3 or 4, this is div(width+3,4)*height; or for
            #  paletteSize of 5 to 16, this is div(width+1,2)*height.
            palette = np.frombuffer(palette_data, dtype=np.uint8).reshape((-1, 3))

            if palette_size > 16:
                # No palette reuse in zrle
                assert palette_size < 127
                bits_per_packed_pixel = 8
            elif palette_size > 4:
                bits_per_packed_pixel = 4
            elif palette_size > 2:
                bits_per_packed_pixel = 2
            else:
                bits_per_packed_pixel = 1

            for j in range(tile_height):
                # Discard any leftover bits for each new line
                b = 0
                nbits = 0

                for i in range(tile_width):
                    if nbits == 0:
                        (b,) = struct.unpack('!B', buf.read(1))
                        nbits = 8
                    nbits -= bits_per_packed_pixel
                    idx = (b >> nbits) & ((1 << bits_per_packed_pixel) - 1) & 127
                    tile[j, i, :] = palette[idx]
            return
        elif run_length_encoded and palette_size == 0:
            # 128:  Plain RLE.  The data consists of a number of runs, repeated
            # until the tile is done.  Runs may continue from the end of one row
            # to the beginning of the next.  Each run is represented by a single
            # pixel value followed by the length of the run.  The length is
            # represented as one or more bytes.  The length is calculated as one
            # more than the sum of all the bytes representing the length.  Any
            # byte value other than 255 indicates the final byte.  So for
            # example, length 1 is represented as [0], 255 as [254], 256 as
            # [255,0], 257 as [255,1], 510 as [255,254], 511 as [255,255,0], and
            # so on.
            #
            # +-------------------------+--------------+-----------------------+
            # | No. of bytes            | Type [Value] | Description           |
            # +-------------------------+--------------+-----------------------+
            # | bytesPerCPixel          | CPIXEL       | pixelValue            |
            # | div(runLength - 1, 255) | U8 array     | 255                   |
            # | 1                       | U8           | (runLength-1) mod 255 |
            # +-------------------------+--------------+-----------------------+
            i = 0
            pixels = tile_width * tile_height
            data = np.zeros((pixels, 3))
            while i < pixels:
                pix = buf.read(bytes_per_pixel)
                pix = np.frombuffer(pix, dtype=np.uint8).reshape((3, ))

                count = 1
                b = None

                while b == 255 or b is None:
                    (b,) = struct.unpack('!B', buf.read(1))
                    count += b

                data[i:i+count, :] = pix
                i += count
            assert i == pixels
        elif run_length_encoded and palette_size > 1:
            # 130 to 255:  Palette RLE.  Followed by the palette, consisting of
            # paletteSize = (subencoding - 128) pixel values:
            #
            #   +----------------------------+--------------+-------------+
            #   | No. of bytes               | Type [Value] | Description |
            #   +----------------------------+--------------+-------------+
            #   | paletteSize*bytesPerCPixel | CPIXEL array | palette     |
            #   +----------------------------+--------------+-------------+
            #
            # Following the palette is, as with plain RLE, a number of runs,
            # repeated until the tile is done.  A run of length one is
            # represented simply by a palette index:
            #
            #         +--------------+--------------+--------------+
            #         | No. of bytes | Type [Value] | Description  |
            #         +--------------+--------------+--------------+
            #         | 1            | U8           | paletteIndex |
            #         +--------------+--------------+--------------+
            #
            # A run of length more than one is represented by a palette index
            # with the top bit set, followed by the length of the run as for
            # plain RLE.
            #
            # +-------------------------+--------------+-----------------------+
            # | No. of bytes            | Type [Value] | Description           |
            # +-------------------------+--------------+-----------------------+
            # | 1                       | U8           | paletteIndex + 128    |
            # | div(runLength - 1, 255) | U8 array     | 255                   |
            # | 1                       | U8           | (runLength-1) mod 255 |
            # +-------------------------+--------------+-----------------------+
            palette = np.frombuffer(palette_data, dtype=np.uint8).reshape((-1, 3))

            i = 0
            pixels = tile_width * tile_height
            data = np.zeros((pixels, 3))
            while i < pixels:
                (idx,) = struct.unpack('!B', buf.read(1))

                count = 1

                if idx & 128:
                    b = None
                    while b == 255 or b is None:
                        (b,) = struct.unpack('!B', buf.read(1))
                        count += b

                idx &= 127
                pix = palette[idx]

                data[i:i+count, :] = pix
                i += count
            assert i == pixels
        else:
            assert False, "Unhandled case: run_length_encoded={} palette_size={}".format(run_length_encoded, palette_size)

        tile[:] = data.reshape((tile_height, tile_width, 3))



from universe.envs import vnc_env
from universe.spaces.joystick_action_space import JoystickActionSpace


class GTAVEnv(vnc_env.VNCEnv):
    def __init__(self):
        super(GTAVEnv, self).__init__()
        self.action_space = JoystickActionSpace(axis_x=True, axis_z=True)
        self._send_actions_over_websockets = True
        self._skip_network_calibration = True




from setuptools import setup, find_packages

setup(name='universe',
      version='0.21.0',
      description="Universe: a software platform for measuring and training an AI's general intelligence across the world's supply of games, websites and other applications.",
      url='https://github.com/openai/universe',
      author='OpenAI',
      author_email='universe@openai.com',
      packages=[package for package in find_packages()
                if package.startswith('universe')],
      install_requires=[
          'autobahn',
          'docker-py==1.10.3',
          'docker-pycreds==0.2.1',
          'fastzbarlight>=0.0.13',
          'go-vncdriver>=0.4.8',
          'gym>=0.6.0',
          'Pillow',
          'PyYAML',
          'six',
          'twisted',
          'ujson',
      ],
      package_data={'universe': ['runtimes.yml']},
      tests_require=['pytest'],
      extras_require={
          'atari': 'gym[atari]',
      }
      )



import logging
from universe import pyprofile
import time
import ujson

from autobahn.twisted import websocket
from twisted.internet import defer

from universe import error

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

class RemoteError(error.Error):
    pass

class RewarderClient(websocket.WebSocketClientProtocol):
    def __init__(self):
        super(RewarderClient, self).__init__()
        self._closed = False
        self._close_message = None

        self._connected = False
        self._requests = {}

        self._reset = None
        self._initial_reset = False

        self._connection_result = defer.Deferred()

    def send_reset(self, env_id, seed, fps, episode_id):
        self._initial_reset = True
        self._reset = {
            'env_id': env_id,
            'fps': fps,
            'episode_id': episode_id,
        }

        return self.send('v0.env.reset', {
            'seed': seed,
            'env_id': env_id,
            'fps': fps,
        }, {'episode_id': episode_id}, expect_reply=True)

    def _finish_reset(self, episode_id):
        extra_logger.info('[%s] Running finish_reset: %s', self.factory.label, episode_id)
        self.reward_buffer.reset(episode_id)

    def onConnect(self, request):
        self._message_id = 0
        self._requests = {}

        self.reward_buffer = self.factory.reward_buffer

        assert not self._connection_result.called
        self._connection_result.callback(self)
        self._connected = True

    def waitForWebsocketConnection(self):
        return self._connection_result

    def send(self, method, body, headers=None, expect_reply=False):
        if headers is None:
            headers = {}
        if self._closed:
            error_message = "Can't send message to closed connection"
            if self._close_message:
                error_message += ": {}".format(self._close_message)
            e = error.Error(error_message)
            if expect_reply:
                return defer.fail(e)
            else:
                raise e

        id = self._message_id

        self._message_id += 1
        new_headers = {
            'message_id': id,
            'sent_at': time.time(),
        }
        new_headers.update(headers)

        payload = {
            'method': method,
            'body': body,
            'headers': new_headers,
        }

        extra_logger.info('[%s] Sending message to rewarder: %s', self.factory.label, payload)
        self.sendMessage(ujson.dumps(payload).encode('utf-8'), False)

        if expect_reply:
            d = defer.Deferred()
            self._requests[id] = (payload, d)
            return d
        else:
            return None

    def _manual_recv(self, method, body, headers={}):
        """Used in the tests"""
        headers.setdefault('sent_at', time.time())
        return self.recv(self._make_context(), {'method': method, 'body': body, 'headers': headers})

    def recv(self, context, response):
        method = response['method']
        body = response['body']
        headers = response['headers']

        remote_time = headers['sent_at']
        local_time = context['start']

        episode_id = headers.get('episode_id')
        if episode_id is not None:
            self.reward_buffer.push_time(episode_id, remote_time, local_time)

        # Gets called by RewarderClient
        if method == 'v0.env.reward':
            episode_id = headers['episode_id']
            reward = body['reward']
            done = body['done']
            info = body['info']
            extra_logger.debug('[%s] Received %s: reward=%s done=%s info=%s episode_id=%s', self.factory.label, method, reward, done, info, episode_id)
            pyprofile.incr('rewarder_client.reward', reward)
            if done:
                pyprofile.incr('rewarder_client.done')
            self.reward_buffer.push(episode_id, reward, done, info)
        elif method == 'v0.env.text':
            episode_id = headers['episode_id']
            text = body['text']
            extra_logger.debug('[%s] Received %s: text=%s episode_id=%s', self.factory.label, method, text, episode_id)
            self.reward_buffer.push_text(episode_id, text)
        elif method == 'v0.env.observation':
            episode_id = headers['episode_id']
            jsonable = body['observation']
            extra_logger.debug('[%s] Received %s: observation=%s episode_id=%s', self.factory.label, method, jsonable, episode_id)
            self.reward_buffer.set_observation(episode_id=episode_id, observation=jsonable)
        elif method == 'v0.env.describe':
            episode_id = headers['episode_id']
            env_id = body['env_id']
            env_state = body['env_state']
            fps = body['fps']
            extra_logger.info('[%s] Received %s: env_id=%s env_state=%s episode_id=%s',
                              self.factory.label, method, env_id, env_state, episode_id)
            self.reward_buffer.set_env_info(env_state, env_id=env_id, episode_id=episode_id, fps=fps)
        elif method == 'v0.reply.env.reset':
            episode_id = headers['episode_id']
            self._finish_reset(episode_id)
        elif method in ['v0.reply.error', 'v0.reply.control.ping']:
            assert headers.get('parent_message_id') is not None
        elif method == 'v0.connection.close':
            assert headers.get('parent_message_id') is None
            logger.debug('Server hanging up: %s', body['message'])

            self._close_message = body['message']
            e = error.Error(body['message'])
            self.factory.record_error(e)
        else:
            logger.error('Unrecognized websocket method: method=%s body=%s headers=%s (consider adding to rewarder_state.py)', method, body, headers)
            return

        parent_id = headers.get('parent_message_id')
        if parent_id is not None:
            try:
                spec = self._requests.pop(parent_id)
            except KeyError:
                logger.error('[%s] Received extra reply to %d; ignoring: method=%s body=%s headers=%s ', self.factory.label, parent_id, method, body, headers)
            else:
                request, d = spec
                if method != 'v0.reply.error':
                    d.callback((context, request, response))
                else:
                    e = RemoteError('[{}] Remote error: {}'.format(self.factory.label, body['message']))
                    d.errback(e)

    def _make_context(self):
        return {'start': time.time()}

    def onMessage(self, payload, isBinary):
        extra_logger.debug('[%s] Received payload: %s', self.factory.label, payload)
        assert not isBinary
        payload = ujson.loads(payload)

        context = self._make_context()
        latency = context['start'] - payload['headers']['sent_at']
        pyprofile.incr('rewarder_protocol.messages')
        pyprofile.incr('rewarder_protocol.messages.{}'.format(payload['method']))

        # Double latency to model RTT
        pyprofile.timing('rewarder_protocol.latency.rtt.skew_unadjusted', 2*latency)
        if latency < 0:
            pyprofile.incr('rewarder_protocol.latency.rtt.skew_unadjusted.negative')

        self.recv(context, payload)

    def onClose(self, wasClean, code, reason):
        if self._close_message:
            return

        if not self._connected:
            assert not self._connection_result.called
            self._connection_result.errback(error.ConnectionError(reason))
            return

        if not self._closed:
            error_message = 'Lost connection: {} (clean={} code={})'.format(reason, wasClean, code)
            reason = error.Error(error_message)
            # TODO: it's not an error if we requested it
            self.factory.record_error(reason)
        else:
            reason = error.Error("We closed the connection: {}".format(reason))

        for request, d in self._requests.values():
            d.errback(reason)

    def close(self, code=1000, reason=None):
        self._closed = True
        extra_logger.info('[%s] Client closing websocket connection because of call to close(code=%s, reason=%s)', self.factory.label, code, reason)
        self.sendClose(code, reason)
        self.transport.loseConnection()



import logging
import gym
import time
import numpy as np

from gym import spaces
from universe import rewarder, vectorized
from universe.wrappers import action_space
from universe.envs.vnc_core_env import translator

logger = logging.getLogger(__name__)

ATARI_HEIGHT = 210
ATARI_WIDTH = 160

class CropAtari(vectorized.ObservationWrapper):
    """
Crop the relevant portion of the monitor where an Atari enviroment resides.
"""

    def __init__(self, env):
        super(CropAtari, self).__init__(env)
        self.observation_space = spaces.Box(0, 255, shape=(ATARI_HEIGHT, ATARI_WIDTH, 3))

    def _observation(self, observation_n):
        return [{'vision': ob['vision'][:ATARI_HEIGHT, :ATARI_WIDTH, :]} for ob in observation_n]

def one_hot(indices, depth):
    return np.eye(depth)[indices]

class GymCoreAction(vectorized.ActionWrapper):
    def __init__(self, env, gym_core_id=None):
        super(GymCoreAction, self).__init__(env)

        if gym_core_id is None:
            # self.spec is None while inside of the make, so we need
            # to pass gym_core_id in explicitly there. This case will
            # be hit when instantiating by hand.
            gym_core_id = self.spec._kwargs['gym_core_id']

        spec = gym.spec(gym_core_id)
        raw_action_space = action_space.gym_core_action_space(gym_core_id)

        self._actions = raw_action_space.actions
        self.action_space = spaces.Discrete(len(self._actions))

        if spec._entry_point.startswith('gym.envs.atari:'):
            self.key_state = translator.AtariKeyState(gym.make(gym_core_id))
        else:
            self.key_state = None

    def _action(self, action_n):
        # Each action might be a length-1 np.array. Cast to int to
        # avoid warnings.
        return [self._actions[int(action)] for action in action_n]

    def _reverse_action(self, action_n):
        # Only works for core envs currently
        self.key_state.apply_vnc_actions(action_n)
        return one_hot(self.key_state.to_index(), self.action_space.n)

class GymCoreObservation(vectorized.Wrapper):
    def __init__(self, env, gym_core_id=None):
        super(GymCoreObservation, self).__init__(env)

        if gym_core_id is None:
            # self.spec is None while inside of the make, so we need
            # to pass gym_core_id in explicitly there. This case will
            # be hit when instantiating by hand.
            gym_core_id = self.spec._kwargs['gym_core_id']

        self._reward_n = None
        self._done_n = None
        self._info_n = None

        self._gym_core_env = gym.spec(gym_core_id).make()

    def _reset(self):
        observation_n = self.env.reset()
        self.reward_n = [0] * self.n
        self.done_n = [False] * self.n
        self.info = {'n': [{} for _ in range(self.n)]}
        new_observation_n, new_reward_n, new_done_n, new_info = self.env.step([[] for i in range(self.n)])
        rewarder.merge_n(
            observation_n, self.reward_n, self.done_n, self.info,
            new_observation_n, new_reward_n, new_done_n, new_info
        )
        return self._observation(self.done_n, self.info)

    def _step(self, action_n):
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        if self.reward_n is not None:
            rewarder.merge_n(
                observation_n, reward_n, done_n, info,
                [None] * self.n, self.reward_n, self.done_n, self.info,
            )
            self.reward_n = self.done_n = self.info = None
        return self._observation(done_n, info), reward_n, done_n, info

    def _observation(self, done_n, info):
        missing = set()

        observation_n = [None] * self.n
        for i, (done, info_i) in enumerate(zip(done_n, info['n'])):
            rewarder_observation = info_i.pop('rewarder.observation', None)
            if rewarder_observation is not None:
                observation, episode_id = rewarder_observation
                observation_n[i] = self._gym_core_env.observation_space.from_jsonable(observation)

                if done:
                    # Check whether we should mask
                    completed = info_i['env_status.completed_episode_id']
                    # Observation from old!
                    if episode_id == completed:
                        logger.debug('[%d] Masking rewarder_observation on episode boundary', i)
                        observation_n[i] = None
            else:
                missing.add(i)

        if len(missing) > 0:
            logger.debug('Missing rewarder observations: %s', missing)
        return observation_n



import gym
from universe import envs, spaces
from universe.wrappers import gym_core_sync

from universe.wrappers.action_space import SafeActionSpace
from universe.wrappers.gym_core import GymCoreAction, GymCoreObservation, CropAtari
from universe.wrappers.blocking_reset import BlockingReset
from universe.wrappers.diagnostics import Diagnostics
from universe.wrappers.joint import Joint
from universe.wrappers.logger import Logger
from universe.wrappers.multiprocessing_env import WrappedMultiprocessingEnv, EpisodeID
from universe.wrappers.render import Render
from universe.wrappers.throttle import Throttle
from universe.wrappers.timer import Timer
from universe.wrappers.vectorize import Vectorize, Unvectorize, WeakUnvectorize
from universe.wrappers.vision import Vision
from universe.wrappers.recording import Recording
from universe.wrappers.monitoring import Monitor

def wrap(env):
    return Timer(Render(Throttle(env)))

def WrappedVNCEnv():
    return wrap(envs.VNCEnv())

def WrappedGymCoreEnv(gym_core_id, fps=None, rewarder_observation=False):
    # Don't need to store the ID on the instance; it'll be retrieved
    # directly from the spec
    env = wrap(envs.VNCEnv(fps=fps))
    if rewarder_observation:
        env = GymCoreObservation(env, gym_core_id=gym_core_id)
    return env

def WrappedGymCoreSyncEnv(gym_core_id, fps=60, rewarder_observation=False):
    spec = gym.spec(gym_core_id)
    env = gym_core_sync.GymCoreSync(BlockingReset(wrap(envs.VNCEnv(fps=fps))))
    if rewarder_observation:
        env = GymCoreObservation(env, gym_core_id=gym_core_id)
    elif spec._entry_point.startswith('gym.envs.atari:'):
        env = CropAtari(env)

    return env

def WrappedFlashgamesEnv():
    keysym = spaces.KeyEvent.by_name('`').key
    return wrap(envs.VNCEnv(probe_key=keysym))

def WrappedInternetEnv(*args, **kwargs):
    return wrap(envs.InternetEnv(*args, **kwargs))

def WrappedStarCraftEnv(*args, **kwargs):
    return wrap(envs.StarCraftEnv(*args, **kwargs))

def WrappedGTAVEnv(*args, **kwargs):
    return wrap(envs.GTAVEnv(*args, **kwargs))

def WrappedWorldOfGooEnv(*args, **kwargs):
    return wrap(envs.WorldOfGooEnv(*args, **kwargs))



from universe.remotes.hardcoded_addresses import HardcodedAddresses
from universe.remotes.allocator_remote import AllocatorManager
from universe.remotes.docker_remote import DockerManager
from universe.remotes.build import build



import logging
from universe import vectorized

logger = logging.getLogger(__name__)

class Vision(vectorized.Wrapper):
    """
At present, an observation from a vectorized universe environment returns a list of 
dicts. Each dict contains input data for each modality.  Modalities include 'vision'
and 'text', and it is possible to add other modalities in the future (such as 'audio').

The Vision wrapper extracts the vision modality and discards all others.  This is convenient
when we only care about the visual input.
"""

    def _reset(self):
        observation_n = self.env.reset()
        return [ob['vision'] if ob is not None else ob for ob in observation_n]

    def _step(self, action_n):
        observation_n, reward_n, done_n, info_n = self.env.step(action_n)
        observation_n = [ob['vision'] if ob is not None else ob for ob in observation_n]
        return observation_n, reward_n, done_n, info_n



from gym.benchmarks import scoring
from gym.benchmarks import register_benchmark

register_benchmark(
    id='Atari7VNC-v0',
    scorer=scoring.TotalReward(),
    name='AtariVNC',
    description='7 Atari games, with pixel observations (using universe)',
    tasks=[
        {
            "env_id": "VNCBeamRider-v3",
            "trials": 1,
            "max_timesteps": 10000000
        },
        {
            "env_id": "VNCBreakout-v3",
            "trials": 1,
            "max_timesteps": 10000000
        },
        {
            "env_id": "VNCEnduro-v3",
            "trials": 1,
            "max_timesteps": 10000000
        },
        {
            "env_id": "gym-core.Pong-v3",
            "trials": 1,
            "max_timesteps": 10000000
        },
        {
            "env_id": "VNCQbert-v3",
            "trials": 1,
            "max_timesteps": 10000000
        },
        {
            "env_id": "VNCSeaquest-v3",
            "trials": 1,
            "max_timesteps": 10000000
        },
        {
            "env_id": "VNCSpaceInvaders-v3",
            "trials": 1,
            "max_timesteps": 10000000
        }
    ])

register_benchmark(
    id='FlashRacing-v0',
    scorer=scoring.RewardPerTime(),
    name='FlashRacing',
    description='7 flash racing games, goal is best score per time',
    tasks=[
        {'env_id': 'flashgames.NeonRace-v0',
         'trials': 1,
         'max_timesteps': 5000000,
         'reward_floor':   175.0,
         'reward_ceiling': 1700.0,
        },
        {'env_id': 'flashgames.CoasterRacer-v0',
         'trials': 1,
         'max_timesteps': 5000000,
         'reward_floor':   17.0,
         'reward_ceiling': 400.0,
        },
        {'env_id': 'flashgames.HeatRushUsa-v0',
         'trials': 1,
         'max_timesteps': 5000000,
         'reward_floor':   150.0,
         'reward_ceiling': 700.0,
        },
        {'env_id': 'flashgames.FormulaRacer-v0',
         'trials': 1,
         'max_timesteps': 5000000,
         'reward_floor':  0.27,
         'reward_ceiling': 1.0,
        },
        {'env_id': 'flashgames.DuskDrive-v0',
         'trials': 1,
         'max_timesteps': 5000000,
         'reward_floor':   5000.0,
         'reward_ceiling': 15000.0,
        },
        {'env_id': 'flashgames.SpacePunkRacer-v0',
         'trials': 1,
         'max_timesteps': 5000000,
         'reward_floor':   0.67,
         'reward_ceiling': 2.25,
        },
        {'env_id': 'flashgames.NeonRace2-v0',
         'trials': 1,
         'max_timesteps': 5000000,
         'reward_floor':   0.0,
         'reward_ceiling': 1200.0,
        }
    ])



import universe.envs.vnc_env
from universe.envs.vnc_env import VNCEnv
from universe.envs.vnc_core_env import GymCoreEnv, GymCoreSyncEnv
from universe.envs.vnc_flashgames import FlashgamesEnv
from universe.envs.vnc_internet import InternetEnv
from universe.envs.vnc_starcraft import StarCraftEnv
from universe.envs.vnc_gtav import GTAVEnv
from universe.envs.vnc_wog import WorldOfGooEnv
from universe.envs.dummy_vnc_env import DummyVNCEnv



from universe import rewarder, spaces, vectorized

class BlockingReset(vectorized.Wrapper):
    """
By default, a reset in universe is not a blocking operation.  This 
wrapper changes it. 
"""

    def __init__(self, *args, **kwargs):
        super(BlockingReset, self).__init__(*args, **kwargs)
        self.reward_n = None
        self.done_n = None
        self.info = None

    def _reset(self):
        observation_n = self.env.reset()
        self.reward_n = [0] * self.n
        self.done_n = [False] * self.n
        self.info = {'n': [{} for _ in range(self.n)]}

        while any(ob is None for ob in observation_n):
            action_n = []
            for done in self.done_n:
                if done:
                    # No popping of reward/done. Don't want to merge across episode boundaries.
                    action_n.append([spaces.PeekReward])
                else:
                    action_n.append([])
            new_observation_n, new_reward_n, new_done_n, new_info = self.env.step(action_n)
            rewarder.merge_n(
                observation_n, self.reward_n, self.done_n, self.info,
                new_observation_n, new_reward_n, new_done_n, new_info
            )
        return observation_n

    def _step(self, action_n):
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        if self.reward_n is not None:
            rewarder.merge_n(
                observation_n, reward_n, done_n, info,
                [None] * self.n, self.reward_n, self.done_n, self.info
            )
            self.reward_n = self.done_n = self.info = None

        while any(ob is None for ob in observation_n):
            action_n = []
            for done in done_n:
                if done:
                    # No popping of reward/done. Don't want to merge across episode boundaries.
                    action_n.append([spaces.PeekReward])
                else:
                    action_n.append([])
            new_observation_n, new_reward_n, new_done_n, new_info = self.env.step(action_n)
            rewarder.merge_n(
                observation_n, reward_n, done_n, info,
                new_observation_n, new_reward_n, new_done_n, new_info
            )
        return observation_n, reward_n, done_n, info



# import numpy as np
# from universe import vectorized

# class SimpleEnv(vectorized.Env):
#     def _step(self, action_n):
#         return {'vision': np.zeros((10, 10))}, 10, False, {}


# from universe.envs import vnc_env
import gym
import numpy as np
import os
import universe

from autobahn.twisted import websocket
from PIL import Image

from universe.rewarder import rewarder_client, rewarder_session, reward_buffer
from universe import spaces, wrappers

def get_rewarder_session(env):
    return env.unwrapped.rewarder_session

def get_vnc_session(env):
    return env.unwrapped.vnc_session

def get_rewarder_client(env):
    rewarder_session = get_rewarder_session(env)
    return rewarder_session.connections['0'].rewarder_client

def get_reward_buffer(env):
    rewarder_session = get_rewarder_session(env)
    return rewarder_session.connections['0'].reward_buffer

def setup_module(module):
    universe.configure_logging('-')

class FakeVNCConnection(object):
    def __init__(self, name, address, password, encoding=None, fine_quality_level=None, subsample_level=None, start_timeout=None):
        self.name = name
        self.address = address
        self.password = password
        self.encoding = encoding
        self.fine_quality_level = fine_quality_level
        self.subsample_level = subsample_level
        self.start_timeout = start_timeout

        self._frame = np.array(Image.open(os.path.join(os.path.dirname(__file__), 'dusk-drive.png')))

    def step(self, action):
        info_d = {}
        return self._frame, info_d

    def _to_dict(self):
        return {
            'name': self.name,
            'address': self.address,
            'password': self.password,
            'encoding': self.encoding,
            'fine_quality_level': self.fine_quality_level,
            'subsample_level': self.subsample_level,
            'start_timeout': self.start_timeout,
        }

class FakeVNCSession(object):
    def __init__(self):
        self.connections = {}

    def connect(self, name, **kwargs):
        self.connections[name] = FakeVNCConnection(name=name, **kwargs)

    def close(self, name=None):
        if name is not None:
            del self.connections[name]
        else:
            self.connections = None

    def step(self, action_d):
        observation_d = {}
        info_d = {}
        err_d = {}
        for name, conn in self.connections.items():
            observation, info = conn.step(action_d.get(name, []))
            observation_d[name] = observation
            info_d[name] = info
        return observation_d, info_d, err_d

    def _to_dict(self):
        return {name: conn._to_dict() for name, conn in self.connections.items()}

class FakeRewarderConnection(object):
    def __init__(self, name, address, label, password, env_id=None, seed=None, fps=60,
                 start_timeout=None, observer=False, skip_network_calibration=False):
        self.name = name
        self.address = address
        self.label = label
        self.password = password
        self.env_id = env_id
        self.seed = None
        self.fps = fps
        self.start_timeout = start_timeout
        self.observer = observer
        self.skip_network_calibration = skip_network_calibration

        self.reward_buffer = reward_buffer.RewardBuffer(label=self.label)

        factory = websocket.WebSocketClientFactory('ws://'+address)
        factory.reward_buffer = self.reward_buffer
        factory.label = self.label
        self.rewarder_client = rewarder_client.RewarderClient()
        self.rewarder_client.factory = factory
        self.rewarder_client.onConnect(None)

    def reset(self, seed=None):
        self.seed = seed

    def pop(self, peek=False):
        return self.reward_buffer.pop(peek=peek)

    def _to_dict(self):
        return {
            'name': self.name,
            'address': self.address,
            'label': self.label,
            'password': self.password,
            'env_id': self.env_id,
            'seed': self.seed,
            'fps': self.fps,
            'start_timeout': self.start_timeout,
            'observer': self.observer,
            'skip_network_calibration': self.skip_network_calibration,
        }

class FakeRewarder(object):
    def __init__(self):
        self.connections = {}

    def reset(self, seed=None):
        for conn in self.connections.values():
            conn.reset(seed=seed)

    def connect(self, name, **kwargs):
        self.connections[name] = FakeRewarderConnection(name=name, **kwargs)
        return rewarder_session.Network()

    def close(self, name=None):
        if name is not None:
            del self.connections[name]
        else:
            self.connections = None

    def pop(self, peek_d):
        reward_d = {}
        done_d = {}
        info_d = {}
        err_d = {}
        for name, conn in self.connections.items():
            reward, done, info = conn.pop(peek=peek_d.get(name))
            reward_d[name] = reward
            done_d[name] = done
            info_d[name] = info
        return reward_d, done_d, info_d, err_d

    def _to_dict(self):
        return {name: conn._to_dict() for name, conn in self.connections.items()}

def test_connect():
    env = gym.make('flashgames.DuskDrive-v0')
    env.configure(vnc_driver=FakeVNCSession, rewarder_driver=FakeRewarder, remotes='vnc://example.com:5900+15900')
    vnc_session = get_vnc_session(env)
    rewarder_session = get_rewarder_session(env)

    assert vnc_session._to_dict() == {'0': {'name': '0', 'subsample_level': 2, 'encoding': 'tight', 'fine_quality_level': 50, 'start_timeout': 7, 'address': 'example.com:5900', 'password': 'openai'}}
    assert rewarder_session._to_dict() == {'0': {'start_timeout': 7, 'seed': None, 'name': '0', 'fps': 60, 'address': 'example.com:15900', 'env_id': 'flashgames.DuskDrive-v0', 'password': 'openai', 'skip_network_calibration': False, 'observer': False, 'label': '0:example.com:5900'}}

def test_describe_handling():
    env = gym.make('flashgames.DuskDrive-v0')
    env.configure(vnc_driver=FakeVNCSession, rewarder_driver=FakeRewarder, remotes='vnc://example.com:5900+15900')
    env.reset()

    reward_buffer = get_reward_buffer(env)
    rewarder_client = get_rewarder_client(env)

    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '1'})

    assert reward_buffer._remote_episode_id == '1'
    assert reward_buffer._remote_env_state == 'resetting'
    assert reward_buffer._current_episode_id == None
    assert reward_buffer.reward_state(reward_buffer._current_episode_id)._env_state == None

    rewarder_client._manual_recv('v0.reply.env.reset', {}, {'episode_id': '1'})

    assert reward_buffer._remote_episode_id == '1'
    assert reward_buffer._remote_env_state == 'resetting'
    assert reward_buffer._current_episode_id == '1'
    assert reward_buffer.reward_state(reward_buffer._current_episode_id)._env_state == 'resetting'

def test_vnc_env():
    env = gym.make('flashgames.DuskDrive-v0')
    env = wrappers.Unvectorize(env)
    env.configure(vnc_driver=FakeVNCSession, rewarder_driver=FakeRewarder, remotes='vnc://example.com:5900+15900')
    env.reset()

    rewarder_client = get_rewarder_client(env)

    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '1'})

    observation, reward, done, info = env.step([spaces.KeyEvent.by_name('a', down=True)])
    assert (observation, reward, done, info['env_status.env_state'], info['env_status.episode_id']) == (None, 0, False, None, None)

    rewarder_client._manual_recv('v0.reply.env.reset', {}, {'episode_id': '1'})

    observation, reward, done, info = env.step([spaces.KeyEvent.by_name('a', down=True)])
    assert (observation, reward, done, info['env_status.env_state'], info['env_status.episode_id']) == (None, 0, False, 'resetting', '1')

    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'running', 'fps': 60}, {'episode_id': '1'})

    rewarder_client._manual_recv('v0.env.reward', {'reward': 10, 'done': False, 'info': {}}, {'episode_id': '1'})
    rewarder_client._manual_recv('v0.env.reward', {'reward': 15, 'done': False, 'info': {}}, {'episode_id': '1'})
    rewarder_client._manual_recv('v0.env.reward', {'reward': -3, 'done': False, 'info': {}}, {'episode_id': '1'})

    observation, reward, done, info = env.step([spaces.KeyEvent.by_name('a', down=True)])
    assert sorted(observation.keys()) == ['text', 'vision']
    assert observation['text'] == []
    assert observation['vision'].shape == (768, 1024, 3)
    assert (reward, done, info['env_status.env_state'], info['env_status.episode_id']) == (22, False, 'running', '1')
    assert info['stats.reward.count'] == 3

def test_boundary_simple():
    env = gym.make('flashgames.DuskDrive-v0')
    env = wrappers.Unvectorize(env)
    env.configure(vnc_driver=FakeVNCSession, rewarder_driver=FakeRewarder, remotes='vnc://example.com:5900+15900')
    env.reset()

    rewarder_client = get_rewarder_client(env)
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '1'})
    rewarder_client._manual_recv('v0.reply.env.reset', {}, {'episode_id': '1'})

    rewarder_client._manual_recv('v0.env.reward', {'reward': 1, 'done': False, 'info': {}}, {'episode_id': '1'})
    rewarder_client._manual_recv('v0.env.reward', {'reward': 2, 'done': True, 'info': {}}, {'episode_id': '1'})
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '2'})

    # We have reward of 3 for episode 1, and episode 2 should now be resetting
    observation, reward, done, info = env.step([])
    assert info['mask.masked.observation']
    assert info['mask.masked.action']
    assert (reward, done, info['env_status.env_state'], info['env_status.episode_id']) == (3, True, 'resetting', '2')

def test_boundary_multiple():
    env = gym.make('flashgames.DuskDrive-v0')
    env = wrappers.Unvectorize(env)
    env.configure(vnc_driver=FakeVNCSession, rewarder_driver=FakeRewarder, remotes='vnc://example.com:5900+15900')
    env.reset()

    rewarder_client = get_rewarder_client(env)
    # episode 2
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '2'})
    rewarder_client._manual_recv('v0.reply.env.reset', {}, {'episode_id': '2'})
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'running', 'fps': 60}, {'episode_id': '2'})
    rewarder_client._manual_recv('v0.env.reward', {'reward': 2, 'done': True, 'info': {}}, {'episode_id': '2'})

    # episode 3
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '3'})
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'running', 'fps': 60}, {'episode_id': '3'})
    rewarder_client._manual_recv('v0.env.reward', {'reward': 3, 'done': True, 'info': {}}, {'episode_id': '3'})

    # episode 4
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '4'})
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'running', 'fps': 60}, {'episode_id': '4'})
    rewarder_client._manual_recv('v0.env.reward', {'reward': 4, 'done': False, 'info': {}}, {'episode_id': '4'})

    observation, reward, done, info = env.step([])
    assert not info.get('mask.masked.observation')
    assert not info.get('mask.masked.action')
    assert (reward, done, info['env_status.env_state'], info['env_status.episode_id']) == (2, True, 'running', '4')
    assert (info['env_status.complete.env_state'], info['env_status.complete.episode_id']) == ('running', '2')

    observation, reward, done, info = env.step([])
    assert (reward, done, info['env_status.env_state'], info['env_status.episode_id']) == (4, False, 'running', '4')

def test_peek():
    env = gym.make('flashgames.DuskDrive-v0')
    env = wrappers.Unvectorize(env)
    env.configure(vnc_driver=FakeVNCSession, rewarder_driver=FakeRewarder, remotes='vnc://example.com:5900+15900')
    env.reset()

    rewarder_client = get_rewarder_client(env)
    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '1'})
    rewarder_client._manual_recv('v0.reply.env.reset', {}, {'episode_id': '1'})

    observation, reward, done, info = env.step([spaces.PeekReward])

    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'resetting', 'fps': 60}, {'episode_id': '2'})
    observation, reward, done, info = env.step([spaces.PeekReward])
    assert info['mask.masked.observation']
    assert info['mask.masked.action']
    assert info['env_status.episode_id'] == '1'
    assert info['env_status.env_state'] == 'resetting'
    assert info['env_status.peek.episode_id'] == '2'
    assert info['env_status.peek.env_state'] == 'resetting'

    rewarder_client._manual_recv('v0.env.describe', {'env_id': 'flashgames.DuskDrive-v0', 'env_state': 'running', 'fps': 60}, {'episode_id': '2'})
    observation, reward, done, info = env.step([spaces.PeekReward])
    assert not info.get('mask.masked.observation')
    assert not info.get('mask.masked.action')
    assert info['env_status.episode_id'] == '1'
    assert info['env_status.env_state'] == 'resetting'
    assert info['env_status.peek.episode_id'] == '2'
    assert info['env_status.peek.env_state'] == 'running'



from universe.vectorized import core

class Filter(object):
    def _after_reset(self, observation):
        return observation

    def _after_step(self, observation, reward, done, info):
        return observation, reward, done, info

class VectorizeFilter(core.Wrapper):
    """Vectorizes a Filter written for the non-vectorized case."""

    autovectorize = False

    def __init__(self, env, filter_factory, *args, **kwargs):
        super(VectorizeFilter, self).__init__(env)
        self.filter_factory = filter_factory
        self._args = args
        self._kwargs = kwargs

    def _configure(self, **kwargs):
        super(VectorizeFilter, self)._configure(**kwargs)
        self.filter_n = [self.filter_factory(*self._args, **self._kwargs) for _ in range(self.n)]

    def _reset(self):
        observation_n = self.env.reset()
        observation_n = [filter._after_reset(observation) for filter, observation in zip(self.filter_n, observation_n)]
        return observation_n

    def _step(self, action_n):
        o_n, r_n, d_n, i = self.env.step(action_n)

        observation_n = []
        reward_n = []
        done_n = []
        info = i.copy()
        info['n'] = []
        for filter, observation, reward, done, info_i in zip(self.filter_n, o_n, r_n, d_n, i['n']):
            observation, reward, done, info_i = filter._after_step(observation, reward, done, info_i)
            observation_n.append(observation)
            reward_n.append(reward)
            done_n.append(done)
            info['n'].append(info_i)
        return observation_n, reward_n, done_n, info

    def __str__(self):
        return '<{}[{}]{}>'.format(type(self).__name__, self.filter_factory, self.env)



import gym
import logging
from universe import rewarder, spaces, vectorized

logger = logging.getLogger(__name__)

class GymCoreSync(vectorized.Wrapper):
    """A synchronized version of the core envs. Its semantics should match
    that of the core envs. (By default, observations are pixels from
    the VNC session, but it also supports receiving the normal Gym
    observations over the rewarder protocol.)

    Provided primarily for testing and debugging.
    """

    def __init__(self, env):
        super(GymCoreSync, self).__init__(env)
        self.reward_n = None
        self.done_n = None
        self.info = None

        # Metadata has already been cloned
        self.metadata['semantics.async'] = False

    def _reset(self):
        observation_n = self.env.reset()
        new_observation_n, self.reward_n, self.done_n, self.info = self.env.step([[] for i in range(self.n)])
        rewarder.merge_observation_n(observation_n, new_observation_n)

        # Fast forward until the observation is caught up with the rewarder
        self._flip_past(observation_n, self.reward_n, self.done_n, self.info)

        assert all(r == 0 for r in self.reward_n), "Unexpectedly received rewards during reset phase: {}".format(self.reward_n)
        return observation_n

    def _step(self, action_n):
        # Add C keypress in order to "commit" the action, as
        # interpreted by the remote.
        action_n = [action + [
            spaces.KeyEvent.by_name('c', down=True),
            spaces.KeyEvent.by_name('c', down=False)
        ] for action in action_n]

        observation_n, reward_n, done_n, info = self.env.step(action_n)
        if self.reward_n is not None:
            rewarder.merge_n(
                observation_n, reward_n, done_n, info,
                [None] * self.n, self.reward_n, self.done_n, self.info,
            )
            self.reward_n = self.done_n = self.info = None

        while True:
            count = len([True for info_i in info['n'] if info_i['stats.reward.count'] == 0])
            if count > 0:
                logger.debug('[GymCoreSync] Still waiting on %d envs to receive their post-commit reward', count)
            else:
                break

            new_observation_n, new_reward_n, new_done_n, new_info = self.env.step([[] for i in range(self.n)])
            rewarder.merge_n(
                observation_n, reward_n, done_n, info,
                new_observation_n, new_reward_n, new_done_n, new_info
            )

        assert all(info_i['stats.reward.count'] == 1 for info_i in info['n']), "Expected all stats.reward.counts to be 1: {}".format(info)

        # Fast forward until the observation is caught up with the rewarder
        self._flip_past(observation_n, reward_n, done_n, info)
        return observation_n, reward_n, done_n, info

    def _flip_past(self, observation_n, reward_n, done_n, info):
        # Wait until all observations are past the corresponding reset times
        remote_target_time = [info_i['reward_buffer.remote_time'] for info_i in info['n']]
        while True:
            new_observation_n, new_reward_n, new_done_n, new_info = self.env.step([[] for i in range(self.n)])

            # info_i.get['diagnostics.image_remote_time'] may not exist, for example when an env
            # is resetting. target is a timestamp, thus > 0, so these will count as "need to catch up"
            deltas = [target - info_i.get('diagnostics.image_remote_time', 0) for target, info_i in zip(remote_target_time, new_info['n'])]
            count = len([d for d in deltas if d > 0])

            rewarder.merge_n(
                observation_n, reward_n, done_n, info,
                new_observation_n, new_reward_n, new_done_n, new_info
            )

            if count == 0:
                return
            else:
                logger.debug('[GymCoreSync] Still waiting on %d envs to catch up to their targets: %s', count, deltas)
